# -*- coding: utf-8 -*-
"""
# --------------------------------------------------------
# @Author : panjq
# @E-mail : pan_jinquan@163.com
# @Date   : 2020-06-02 16:00:47
# --------------------------------------------------------
"""
import torch
import random
import os
import numpy as np
from collections import OrderedDict
from collections.abc import Iterable


def get_torch_version():
    try:
        v = torch.__version__
        print("torch.version:{}".format(v))
        vid = torch_version_id(v)
    except Exception as e:
        vid = None
    return vid


def torch_version_id(v: str):
    vid = v.split(".")
    vid = float("{}.{:0=2d}".format(vid[0], int(vid[1])))
    return vid


def set_env_random_seed(seed=2020):
    """
    :param seed:
    :return:
    """
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)  # if you are using multi-GPU.
    # torch.backends.cudnn.benchmark = False
    # torch.backends.cudnn.deterministic = True


def get_loacl_eth():
    '''
    想要获取linux设备网卡接口，并用列表进行保存
    :return:
    '''
    eth_list = []
    cmd = "ls -l /sys/class/net/ | grep -v virtual | sed '1d' | awk 'BEGIN {FS=\"/\"} {print $NF}'"
    try:
        with os.popen(cmd) as f:
            for line in f.readlines():
                line = line.strip()
                eth_list.append(line.lower())
    except Exception as e:
        print(e, "can not found eth,will set default eth is:eth0")
        eth_list = ["eth0"]
    if not eth_list:
        eth_list = ["eth0"]
    return eth_list


def set_node_env(master_addr="localhost", master_port="1200", eth_name=None):
    """
    设置多卡训练的节点信息
    parser = argparse.ArgumentParser(description="for face verification train")
    parser.add_argument("-c", "--config", help="configs file", default="configs/config_distributed.yaml", type=str)
    parser.add_argument("-e", '--eth_name', type=str, default=None, help="set eth name")
    parser.add_argument("-a", '--master_addr', type=str, default='localhost', help="set master node address")
    parser.add_argument("-p", '--master_port', type=str, default='1200', help="set master node port")
    parser.add_argument("--local_rank", type=int, default=0, help="torch.distributed.launch会给模型分配一个args.local_rank的参数，"
                                                                  "也可以通过torch.distributed.get_rank()获取进程id")
    parser.add_argument("--init_method", type=str, default="env://")
    args = parser.parse_args()
    ====================================
    :param master_addr: 主节点地址,default localhost
    :param master_port: 主节点接口,default 1200
    :param eth_name: 网卡名称，None会自动获取
    :return:
    """
    if eth_name is None:  # auto get eth_name
        eth_name = get_loacl_eth()[0]
        print("eth_name:{}".format(eth_name))
    os.environ['NCCL_SOCKET_IFNAME'] = eth_name
    os.environ['MASTER_ADDR'] = master_addr
    os.environ['MASTER_PORT'] = master_port


def set_distributed_env(backend="nccl", init_method="env://"):
    """
    initialize the distributed environment
    :param backend:
    :param init_method:
    :return: world_size :参与工作的进程数
             rank: 当前进程的rank(这个Worker是全局第几个Worker)
             local_rank：这个Worker是这台机器上的第几个Worker
    """
    # initialize process group
    # use nccl backend to speedup gpu communication
    torch.distributed.init_process_group(backend=backend, init_method=init_method)
    world_size = torch.distributed.get_world_size()
    # torch.distributed.launch 会给模型分配一个args.local_rank的参数，也可以通过torch.distributed.get_rank()获取进程id。
    rank = torch.distributed.get_rank()  # os.environ["RANK"]
    return world_size, rank


def get_distributed_sampler(dataset: torch.utils.data.Dataset, world_size, rank):
    """
    Example:
        sampler = DistributedSampler(dataset) if is_distributed else None
        loader = DataLoader(dataset, shuffle=(sampler is None),
                             sampler=sampler)
        for epoch in range(start_epoch, n_epochs):
             if is_distributed:
                 sampler.set_epoch(epoch)
             train(loader)
    :param dataset:
    :param world_size:
    :param rank:
    :return:
    """
    sampler = torch.utils.data.distributed.DistributedSampler(
        dataset, num_replicas=world_size, rank=rank)
    return sampler


def get_device():
    """
    返回当前设备索引
    torch.cuda.current_device()
    返回GPU的数量
    torch.cuda.device_count()
    返回gpu名字，设备索引默认从0开始
    torch.cuda.get_device_name(0)
    cuda是否可用
    torch.cuda.is_available()
    ==========
    CUDA_VISIBLE_DEVICES=4,5,6 python train.py

    Usage:
    gpu_id = get_device()
    model = build_model()
    model.cuda()
    model = torch.nn.DataParallel(model, device_ids=gpu_id)
    ...
    :return:
    """
    gpu_id = list(range(torch.cuda.device_count()))
    return gpu_id


def print_model(model):
    """
    :param model:
    :return:
    """
    for k, v in model.named_parameters():
        # print(k,v)
        print(k)


def freeze_net_layers(net):
    """
    https://www.zhihu.com/question/311095447/answer/589307812
    example:
        freeze_net_layers(net.base_net)
        freeze_net_layers(net.source_layer_add_ons)
        freeze_net_layers(net.extras)
    :param net:
    :return:
    """
    # for param in net.parameters():
    #     param.requires_grad = False
    for name, child in net.named_children():
        # print(name, child)
        for param in child.parameters():
            param.requires_grad = False


def rename_module(state_dict, name_map={}):
    """
    重新映射模块名称
    :param state_dict: model state_dict
    :param name_map: 需要重新命名的模型， name_map = {"base_net": "backbone"},即将base_net重新命名为backbone
    :return:
    """
    # 初始化一个空 dict
    new_state_dict = OrderedDict()
    # 修改 key，没有module字段则需要不del，如果有，则需要修改为 module.features
    for k, v in state_dict.items():
        for key, value in name_map.items():
            if key in k: k = k.replace(key, value)
        new_state_dict[k] = v
    return new_state_dict


def load_state_dict(model_path):
    """
    Usage:
        model=Model()
        state_dict = torch_tools.load_state_dict(model_path, module=False)
        model.load_state_dict(state_dict)
    :param model_path:
    :return: state_dict
    """
    state_dict = None
    if model_path:
        print('=> loading model from {}'.format(model_path))
        state_dict = torch.load(model_path, map_location=torch.device('cpu'))
        if "model" in state_dict:
            state_dict = state_dict["model"]
        if 'module' in list(state_dict.keys())[0]:
            state_dict = get_module_state_dict(state_dict)
    else:
        raise Exception("Error:no model file:{}".format(model_path))
    return state_dict


def get_module_state_dict(state_dict):
    """
    :param state_dict:
    :return:
    """
    # 初始化一个空 dict
    new_state_dict = OrderedDict()
    # 修改 key，没有module字段则需要不del，如果有，则需要修改为 module.features
    for k, v in state_dict.items():
        if k.startswith("module."):
            # k = k.replace('module.', '')
            k = k[len("module."):]
        new_state_dict[k] = v
    return new_state_dict


def load_pretrained_model(model, ckpt):
    """Loads pretrianed weights to model.
    Features:只会加载完全匹配的模型参数，不匹配的模型将会忽略
        - Incompatible layers (unmatched in name or size) will be ignored.
        - Can automatically deal with keys containing "module.".
    Args:
        model (nn.Module): network model.
        ckpt (str): OrderedDict or model file
    """
    if isinstance(ckpt, str):
        checkpoint = load_state_dict(ckpt)
        if 'state_dict' in checkpoint:
            state_dict = checkpoint['state_dict']
        else:
            state_dict = checkpoint
    elif isinstance(ckpt, OrderedDict):
        state_dict = ckpt
    else:
        raise Exception("nonsupport type:{} ".format(ckpt))
    model_dict = model.state_dict()
    new_state_dict = OrderedDict()
    matched_layers, discarded_layers = [], []
    for k, v in state_dict.items():
        if k.startswith('module.'):
            k = k[7:]  # discard module.
        if k in model_dict and model_dict[k].size() == v.size():
            new_state_dict[k] = v
            matched_layers.append(k)
        else:
            discarded_layers.append(k)
    model_dict.update(new_state_dict)
    model.load_state_dict(model_dict)
    print("=" * 60)
    if len(matched_layers) == 0:
        raise Exception('The model checkpoint cannot be loaded,'
                        'please check the key names manually')
    else:
        print('Successfully loaded model checkpoint')
        # [print('{}'.format(layer)) for layer in matched_layers]
        if len(discarded_layers) > 0:
            print('The following layers are discarded due to unmatched keys or layer size')
            [print('{}'.format(layer)) for layer in discarded_layers]

    print("=" * 60)
    return model


def plot_model(model, output=None, input_shape=None):
    """
    Usage:
    output = model(inputs)
    vis_graph = make_dot(output, params=dict(model.named_parameters()))
    vis_graph.view()
    =================================================================
    output/input_shape至少已知一个
    :param model:
    :param output:
    :param input_shape: (batch_size, 3, input_size[0], input_size[1])
    :return:
    """
    from torchviz import make_dot
    if output is None:
        output = model_forward(model, input_shape, device="cpu")
    vis_graph = make_dot(output, params=dict(model.named_parameters()))
    vis_graph.view()


def model_forward(model, input_shape, device="cpu"):
    """
    input_shape=(batch_size, 3, input_size[0], input_size[1])
    :param model:
    :param input_shape:
    :param device:
    :return:
    """
    inputs = torch.randn(size=input_shape)
    inputs = inputs.to(device)
    model = model.to(device)
    model.eval()
    output = model(inputs)
    return output


def summary_model(model, batch_size=1, input_size=[112, 112], plot=False, device="cpu"):
    """
    ----This tools can show----
    Total params: 359,592
    Total memory: 47.32MB
    Total MAdd: 297.37MMAdd
    Total Flops: 153.31MFlops
    Total MemR+W: 99.7MB
    ====================================================
    https://www.cnblogs.com/xuanyuyt/p/12653041.html
    Total number of network parameters (params)
    Theoretical amount of floating point arithmetics (FLOPs)
    Theoretical amount of multiply-adds (MAdd MACC) (乘加运算)
    Memory usage (memory)
    MACCs：是multiply-accumulate operations，指点积运算， 一个 macc = 2FLOPs
    FLOPs 的全称是 floating points of operations，即浮点运算次数，用来衡量模型的计算复杂度。
    计算 FLOPs 实际上是计算模型中乘法和加法的运算次数。
    卷积层的浮点运算次数不仅取决于卷积核的大小和输入输出通道数，还取决于特征图的大小；
    而全连接层的浮点运算次数和参数量是相同的。
    ====================================================
    :param model:
    :param batch_size:
    :param input_size:
    :param plot: plot model
    :param device:
    :return:
    """
    from torchsummary import summary
    from torchstat import stat
    inputs = torch.randn(size=(batch_size, 3, input_size[1], input_size[0]))
    inputs = inputs.to(device)
    model = model.to(device)
    model.eval()
    output = model(inputs)
    # 统计模型参数
    summary(model, input_size=(3, input_size[1], input_size[0]), batch_size=batch_size, device=device)
    # 统计模型参数和计算FLOPs
    stat(model, (3, input_size[1], input_size[0]))
    # summary可能报错，可使用该方法
    # summary_v2(model, inputs, item_length=26, verbose=True)
    # from thop import profile
    # macs, params = profile(model, inputs=(inputs,))
    # print("Total Flops :{}".format(macs))
    # print("Total params:{}".format(params))
    print("===" * 10)
    print("inputs.shape:{}".format(inputs.shape))
    # print("output.shape:{}".format(output.shape))
    if plot:
        plot_model(model, output)


def nni_summary_model(model, batch_size=1, input_size=[112, 112], plot=False, device="cpu"):
    """
    https://nni.readthedocs.io/zh/stable/Compression/CompressionUtils.html
    NNI 提供了模型计数器，用于计算模型的 FLOPs 和参数。
    此计数器支持计算没有掩码模型的 FLOPs、参数，也可以计算有掩码模型的 FLOPs、参数，
    这有助于在模型压缩过程中检查模型的复杂度。
    注意，对于结构化的剪枝，仅根据掩码来标识保留的滤波器，不会考虑剪枝的输入通道，
    因此，计算出的 FLOPs 会比实际数值要大（即，模型加速后的计算值）。
    我们支持两种模式来收集模块信息。 第一种是 default 模式，它只采集卷积操作和线性操作的信息。
    第二种是 full 模式，它还会收集其他操作的信息。 用户可以轻松地使用我们收集的 results 进行进一步的分析。
    :param model:
    :param batch_size:
    :param input_size:
    :param plot:
    :param device:
    :return:
    """
    from nni.compression.pytorch.utils.counter import count_flops_params
    inputs = torch.randn(size=(batch_size, 3, input_size[1], input_size[0]))
    inputs = inputs.to(device)
    model = model.to(device)
    model.eval()
    output = model(inputs)
    flops, params, _ = count_flops_params(model, inputs, verbose=True)
    info = f"Model FLOPs {flops / 1e6:.2f}M, Params {params / 1e6:.2f}M"
    print(info)
    if plot:
        plot_model(model, output)


def torchinfo_summary(model, batch_size=1, input_size=[112, 112], plot=False, device="cpu"):
    """
    ----This tools can show----
    Total params: 359,592
    Total memory: 47.32MB
    Total MAdd: 297.37MMAdd
    Total Flops: 153.31MFlops
    Total MemR+W: 99.7MB
    ====================================================
    https://www.cnblogs.com/xuanyuyt/p/12653041.html
    Total number of network parameters (params)
    Theoretical amount of floating point arithmetics (FLOPs)
    Theoretical amount of multiply-adds (MAdd MACC) (乘加运算)
    Memory usage (memory)
    MACCs：是multiply-accumulate operations，指点积运算， 一个 macc = 2FLOPs
    FLOPs 的全称是 floating points of operations，即浮点运算次数，用来衡量模型的计算复杂度。
    计算 FLOPs 实际上是计算模型中乘法和加法的运算次数。
    卷积层的浮点运算次数不仅取决于卷积核的大小和输入输出通道数，还取决于特征图的大小；
    而全连接层的浮点运算次数和参数量是相同的。
    ====================================================
    :param model:
    :param batch_size:
    :param input_size:
    :param plot: plot model
    :param device:
    :return:
    """
    from torchinfo import summary
    from torchstat import stat
    inputs = torch.randn(size=(batch_size, 3, input_size[1], input_size[0]))
    inputs = inputs.to(device)
    model = model.to(device)
    model.eval()
    output = model(inputs)
    # 统计模型参数
    summary(model, input_size=(batch_size, 3, input_size[1], input_size[0]), device=device)
    # 统计模型参数和计算FLOPs
    stat(model, (3, input_size[1], input_size[0]))
    # summary可能报错，可使用该方法
    # summary_v2(model, inputs, item_length=26, verbose=True)
    # from thop import profile
    # macs, params = profile(model, inputs=(inputs,))
    # print("Total Flops :{}".format(macs))
    # print("Total params:{}".format(params))
    print("===" * 10)
    print("inputs.shape:{}".format(inputs.shape))
    # print("output.shape:{}".format(output.shape))
    if plot:
        plot_model(model, output)



# -*-coding: utf-8 -*-
"""
    @Project: python-learning-notes
    @File   : pandas_tools.py
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2019-07-30 20:13:59
"""
import pandas as pd
import numpy as np
import pandas as pd


def read_csv(filename, usecols=None):
    file = pd.read_csv(filename, usecols=usecols)
    df = pd.DataFrame(file)
    return df


def save_csv(filename, df, save_index=True):
    df.to_csv(filename, index=save_index, sep=',', header=True)


def print_info(class_name, labels):
    # index =range(len(class_name))+1
    index = np.arange(0, len(class_name)) + 1
    columns = ['class_name', 'labels']
    content = np.array([class_name, labels]).T
    df = pd.DataFrame(content, index=index, columns=columns)  # 生成6行4列位置
    print(df)  # 输出6行4列的表格
    save_csv("my_test.csv", df)


def construct_pd(index, columns_name, content, filename=None):
    df = pd.DataFrame(content, index=index, columns=columns_name)  # 生成6行4列位置
    if filename is not None:
        save_csv(filename, df)
    return df


if __name__ == "__main__":
    class_name = ['C1', 'C2', 'C3']
    labels = [100, 200, 300]
    # print_info(class_name, labels)
    index = [1, 2, 3, 4, 5, 8]
    columns_name = ["A", "B"]
    content = np.arange(0, 12).reshape(6, 2)
    filename = "my_test.csv"
    df = construct_pd(index, columns_name, content, filename)
    print(df)



# -*-coding: utf-8 -*-
"""
    @Project: utils
    @File   : pr.py
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2019-05-10 17:10:56
"""
# import matplotlib
import matplotlib.pyplot as plt
from sklearn import metrics
import numpy as np
import sklearn.model_selection as cross_validation


def get_precision(true_labels, pred_labels, average='binary'):
    '''
    https://blog.csdn.net/CherDW/article/details/55813071
    see @get_precision_recall
    precision=tp / (tp + fp)
    :param true_labels:
    :param pred_labels:
    :return:
    '''
    precision = metrics.precision_score(true_labels, pred_labels, average=average)
    return precision


def get_recall(true_labels, pred_labels, average="binary"):
    '''
    https://blog.csdn.net/CherDW/article/details/55813071
    see @get_precision_recall
    recall=tp / (tp + fn)
    :param true_labels:
    :param pred_labels:
    :param  average : string, [None, 'binary' (default), 'micro', 'macro', 'samples', weighted']
    :return:
    '''
    recall = metrics.recall_score(true_labels, pred_labels, average=average)
    return recall


def get_accuracy(true_labels, pred_labels):
    accuracy = metrics.accuracy_score(true_labels, pred_labels)
    return accuracy


def get_precision_recall_acc(true_labels, pred_labels, average="binary"):
    '''
    将一个二分类matrics拓展到多分类或多标签问题时，我们可以将数据看成多个二分类问题的集合，
    每个类都是一个二分类。接着，我们可以通过跨多个分类计算每个二分类metrics得分的均值，
    这在一些情况下很有用。你可以使用average参数来指定。
    binary :二分类
    macro：计算二分类metrics的均值，为每个类给出相同权重的分值。当小类很重要时会出问题，
          因为该macro-averging方法是对性能的平均。另一方面，该方法假设所有分类都是一样重要的，
          因此macro-averaging方法会对小类的性能影响很大。
    weighted:对于不均衡数量的类来说，计算二分类metrics的平均，通过在每个类的score上进行加权实现。
    micro：给出了每个样本类以及它对整个metrics的贡献的pair（sample-weight），
          而非对整个类的metrics求和，它会每个类的metrics上的权重及因子进行求和，来计算整个份额。
          Micro-averaging方法在多标签（multilabel）问题中设置，包含多分类，此时，大类将被忽略。
    samples：应用在multilabel问题上。它不会计算每个类，相反，它会在评估数据中，
          通过计算真实类和预测类的差异的metrics，来求平均（sample_weight-weighted）
    average：average=None将返回一个数组，它包含了每个类的得分.
    原文：https://blog.csdn.net/CherDW/article/details/55813071
    :param true_labels:
    :param pred_labels:
    :return:
    '''
    precision = get_precision(true_labels, pred_labels, average)
    recall = get_recall(true_labels, pred_labels, average)
    acc = get_accuracy(true_labels, pred_labels)
    return precision, recall, acc


def plot_pr_curve(recall_list, precision_list, auc_list, line_names):
    '''
    绘制roc曲线
    :param recall_list:
    :param precision_list:
    :param auc_list:
    :param line_names:曲线名称
    :return:
    '''
    # 绘图
    # plt.figure()
    lw = 2
    plt.figure(figsize=(10, 10))
    colors = ["b", "r", "c", "m", "g", "lt_steps", "k", "w"]
    for r, p, roc_auc, color, line_name in zip(recall_list, precision_list, auc_list, colors, line_names):
        plt.plot(r, p, color=color, lw=lw,
                 label='{} PR curve (area = {:.3f})'.format(line_name, roc_auc))  # 假正率为横坐标，真正率为纵坐标做曲线
    # plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')

    plt.plot([0, 1], [1, 0], color='navy', lw=lw, linestyle='--')  # 绘制y=1-x的直线

    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    # 设置横纵坐标的名称以及对应字体格式
    font = {'family': 'Times New Roman',
            'weight': 'normal',
            'size': 20,
            }
    plt.xlabel('Recall', font)
    plt.ylabel('Precision', font)

    plt.title('PR curve')
    plt.legend(loc="lower right")  # "upper right"
    # plt.legend(loc="upper right")#"upper right"
    plt.show()


def custom_pr_curve(y, pred):
    lw = 2
    plt.figure(figsize=(10, 10))

    pos = np.sum(y == 1)
    neg = np.sum(y == 0)
    pred_sort = np.sort(pred)[::-1]  # 从大到小排序
    index = np.argsort(pred)[::-1]  # 从大到小排序
    y_sort = y[index]
    print(y_sort)

    Pre = []
    Rec = []
    for i, item in enumerate(pred_sort):
        if i == 0:  # 因为计算precision的时候分母要用到i，当i为0时会出错，所以单独列出
            Pre.append(1)
            Rec.append(0)
        else:
            Pre.append(np.sum((y_sort[:i] == 1)) / i)
            Rec.append(np.sum((y_sort[:i] == 1)) / pos)
    print(Pre)
    print(Rec)
    # 画图
    plt.plot(Rec, Pre, 'k')
    plt.plot([0, 1], [1, 0], color='navy', lw=lw, linestyle='--')  # 绘制y=1-x的直线
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    # 设置横纵坐标的名称以及对应字体格式
    font = {'family': 'Times New Roman',
            'weight': 'normal',
            'size': 20,
            }
    plt.xlabel('Recall', font)
    plt.ylabel('Precision', font)

    plt.title('PR curve')
    plt.legend(loc="lower right")  # "upper right"
    # plt.legend(loc="upper right")#"upper right"
    plt.show()


def get_classification_precision_recall(y_true, probas_pred):
    '''
    对于二分类问题，可以直接调用sklearn.metrics的precision_recall_curve()
    :param y_true: 真实样本的的正负标签
    :param probas_pred: 预测的分数
    precision : array, shape = [n_thresholds + 1]
        Precision values such that element i is the precision of
        predictions with score >= thresholds[i] and the last element is 1.

    recall : array, shape = [n_thresholds + 1]
        Decreasing recall values such that element i is the recall of
        predictions with score >= thresholds[i] and the last element is 0.
    :return:
    '''

    precision, recall, _thresholds = metrics.precision_recall_curve(y_true, probas_pred)
    AUC = metrics.auc(recall, precision)
    return precision, recall, AUC


def get_object_detection_precision_recall(iou_data, probas_pred, iou_threshold):
    '''
    在二分类中，recall的分母是测试样本的的正样本的个数，因此可认为二分类recall是召回正样本的比率
    sklearn.metrics的precision_recall_curve()是根据二分类返回的recall；
    但对于目标检测问题，召回率recall与二分类的情况不同，目标检测的召回率recall的分母是groundtruth boundingbox的个数
    可以认为目标检测的recall是召回groundtruth boundingbox的比率；
    显示将precision_recall_curve()返回recall*正样本的个数/groundtruth boundingbox个数就是目标检测的recall了
    :param iou_data:
    :param probas_pred:
    :param iou_threshold:
    :return:
    '''
    true_label = np.where(iou_data > iou_threshold, 1, 0)
    t_nums = np.sum(true_label)  # 测试样本中正样本个数
    gb_nums = len(iou_data)  # groundtruth boundingbox的个数
    precision, recall, _thresholds = metrics.precision_recall_curve(true_label, probas_pred)
    recall = recall * t_nums / gb_nums
    AUC = metrics.auc(recall, precision)
    return precision, recall, AUC


def plot_classification_pr_curve(true_label, prob_data, invert=False, plot_pr=True):
    if invert:
        true_label = 1 - true_label  # 当y_test与y_score是反比关系时,进行反转
    precision, recall, AUC = get_classification_precision_recall(y_true=true_label, probas_pred=prob_data)
    # 绘制ROC曲线
    if plot_pr:
        # custom_pr_curve(true_label, prob_data)
        recall_list = [recall]
        precision_list = [precision]
        auc_list = [AUC]
        line_names = ["line_name"]
        plot_pr_curve(recall_list, precision_list, auc_list, line_names)
    return precision, recall, AUC


def plot_object_detection_pr_curve(iou_data, prob_data, iou_threshold, plot_pr=True):
    precision, recall, AUC = get_object_detection_precision_recall(iou_data=iou_data, probas_pred=prob_data,
                                                                   iou_threshold=iou_threshold)
    # 绘制ROC曲线
    if plot_pr:
        recall_list = [recall]
        precision_list = [precision]
        auc_list = [AUC]
        line_names = [""]
        plot_pr_curve(recall_list, precision_list, auc_list, line_names)


def precision_recall_demo():
    true_labels = [0, 1, 2, 3, 4, 5]
    pred_labels = [0, 2, 1, 0, 0, 1]
    # true_labels = [0, 1, 3, 2, 1]
    # pred_labels = [0, 1, 1, 0, 0]
    precision, recall, acc = get_precision_recall_acc(true_labels, pred_labels, average="macro")
    print("precision:{},recall:{},acc:{}".format(precision, recall, acc))


if __name__ == "__main__":
    iou_data = [0.88, 0.4, 0.70]
    prob_data = [0.9, 0.8, 0.7]
    iou_data = np.array(iou_data)
    prob_data = np.array(prob_data)
    # 阈值
    iou_threshold = 0.5
    true_label = np.where(iou_data > iou_threshold, 1, 0)
    # plot_classification_pr_curve(true_label, prob_data)
    # plot_object_detection_pr_curve(iou_data, prob_data, iou_threshold, plot_pr=True)

    precision_recall_demo()



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-12-09 19:16:19
"""
import os
import copy
import torch
import torch.nn as nn
import torch.onnx
import torch.nn.functional as F
from nni.compression.pytorch.utils.counter import count_flops_params
from nni.compression.pytorch.speedup import ModelSpeedup
from nni.algorithms.compression.pytorch import pruning
from nni.compression.pytorch import apply_compression_results


def model_pruning(model: nn.Module,
                  input_size=[1, 3, 128, 128],
                  sparsity=0.2,
                  prune_mod="L1",
                  reuse=False,
                  speedup=True,
                  output_prune="pruning_output",
                  mask_file="",
                  dependency_aware=True,
                  device="cpu",
                  verbose=False,
                  **kwargs):
    """
    使用NNI进行模型剪枝和压缩
    https://github.com/microsoft/nni/blob/master/docs/en_US/Compression/compression_pipeline_example.ipynb
    use l1filter pruner to prune the model
    Note that if you use a compressor that need you to pass a optimizer,
    you need a new optimizer instead of you have used above, because NNI might modify the optimizer.
    And of course this modified optimizer can not be used in finetuning.
    Usage:
        model = build_model()
        model = model_pruning(model,input_size,sparsity=0.2)
        optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
        for epoch in range(0, epochs):
            trainer(model, optimizer, criterion, epoch)
            evaluator(model)
            torch.save(model.state_dict(), "model_pruning.pth")
    :param model:  Pytorch模型
    :param input_size: 模型输入的维度[batch-size,channel,H,W]
    :param sparsity: 模型剪枝目标稀疏度,值越大,模型越稀疏，计算量越小；0.5表示剪除50%
    :param reuse: 是否复现模型剪枝的结果
                False: 进行模型剪枝(pruning+ModelSpeedup),会生成mask文件(mask_naive_l1filter.pth)
                True : 复现模型剪枝的结果,需要提供mask_file文件
    :param speedup: 是否加速模型
    :param output_prune: 模型剪枝输出文件
    :param mask_file: reuse=True需要提供模型剪枝的mask文件,默认保存在output_prune目录下(mask_naive_l1filter.pth)
    :param dependency_aware 依赖感知模式 https://nni.readthedocs.io/zh/stable/Compression/DependencyAware.html
    :param device:
    :param config: 模型剪枝配置,用于指定需要剪枝网络层；
                        如果不指定op_names,默认对所以层进行剪枝
                        config_list = [{'sparsity': sparsity,'op_types': ['Conv2d'],'op_names': ['conv1']}]
    :return:
    """
    info = ""
    model = model.to(device)
    if not os.path.exists(output_prune): os.makedirs(output_prune)
    prune_file = os.path.join(output_prune, 'pruned_naive_{}filter.pth'.format(prune_mod))
    onnx_file = os.path.join(output_prune, 'pruned_naive_{}filter.onnx'.format(prune_mod))
    mask_file = os.path.join(output_prune, 'mask_naive_{}filter.pth'.format(prune_mod)) if not mask_file else mask_file
    dummy_input = torch.randn(input_size).to(device)
    # 原始模型的计算量和参数量
    flops, params, _ = count_flops_params(model, dummy_input, verbose=verbose)
    info += f"origin-Model FLOPs {flops / 1e6:.2f}M, Params {params / 1e6:.2f}M\n"
    # 模型剪枝,会生成mask文件(mask_naive_l1filter.pth)
    if not reuse:
        """
        Choose a pruner and pruning 
        use l1filter pruner to prune the model
        Note that if you use a compressor that need you to pass a optimizer,
        you need a new optimizer instead of you have used above, because NNI might modify the optimizer.
        And of course this modified optimizer can not be used in finetuning.
        """
        if prune_mod.lower() == "Level".lower():
            config = [{'sparsity': sparsity, 'op_types': ['Conv2d']}]
            pruner = pruning.LevelPruner(model, config)
        elif prune_mod.lower() == "L1".lower():
            # op_types : Only Conv2d is supported in L1FilterPruner.
            # config = [{'sparsity': sparsity, 'op_types': ['Conv2d'], "exclude": False}]
            config = [{'sparsity': sparsity, 'op_types': ['Conv2d']}]
            pruner = pruning.L1FilterPruner(model, config, dependency_aware, dummy_input=dummy_input)
        elif prune_mod.lower() == "L2".lower():
            # op_types : Only Conv2d is supported in L2FilterPruner.
            config = [{'sparsity': sparsity, 'op_types': ['Conv2d']}]
            pruner = pruning.L2FilterPruner(model, config, dependency_aware, dummy_input=dummy_input)
        elif prune_mod.lower() == "FPGM".lower():
            # op_types : Only Conv2d is supported in FPGM Pruner
            config = [{'sparsity': sparsity, 'op_types': ['Conv2d']}]
            pruner = pruning.FPGMPruner(model, config, dependency_aware, dummy_input=dummy_input)
        elif prune_mod.lower() == "Slim".lower():
            # op_types : Only BatchNorm2d is supported in Slim Pruner.
            config = [{'sparsity': sparsity, 'op_types': ['BatchNorm2d']}]
            pruner = pruning.SlimPruner(model,
                                        config,
                                        optimizer=None,
                                        trainer=None,
                                        criterion=None,
                                        sparsifying_training_epochs=10)
        elif prune_mod.lower() == "Slim".lower():
            config = [{'sparsity': sparsity, 'op_types': ['BatchNorm2d']}]
            pruner = pruning.ActivationMeanRankFilterPruner()

        else:
            raise Exception("Error prune_mod:{}".format(prune_mod))
        # compress the model, the mask will be updated.
        pruner.compress()
        # pruner.get_pruned_weights()
        # use a dummy input to apply the sparsify.
        out = model(dummy_input)
        # 剪枝后模型的计算量和参数量
        flops, params, _ = count_flops_params(model, dummy_input, verbose=verbose)
        info += f"pruner-Model FLOPs {flops / 1e6:.2f}M, Params {params / 1e6:.2f}M\n"
        # export the sparsified and mask model
        pruner.export_model(model_path=prune_file, mask_path=mask_file,
                            onnx_path=onnx_file, input_shape=dummy_input.shape,
                            device=device,
                            opset_version=11)
        # speedup the model with provided weight mask.If you use a wrapped model, don't forget to unwrap it.
        pruner._unwrap_model()
        # 将掩码应用到模型,模型会变得更小，推理延迟也会减小
        # apply_compression_results(model, mask_file, device)
    if speedup:
        if not os.path.exists(mask_file): raise Exception("not found mask file:{}".format(mask_file))
        print("load mask file to speed up:{}".format(mask_file))
        speed_up = ModelSpeedup(model, dummy_input=dummy_input, masks_file=mask_file)
        speed_up.speedup_model()
        out = model(dummy_input)
        # speedup后模型的计算量和参数量
        flops, params, _ = count_flops_params(model, dummy_input, verbose=verbose)
        info += f"speedup-Model FLOPs {flops / 1e6:.2f}M, Params {params / 1e6:.2f}M\n"
    # finetune the model to recover the accuracy.
    model = model.to(device)
    print(info)
    return model


def nni_model_pruning_test(model, input_size=[1, 3, 416, 416], sparsity=0.5, output="output", device="cpu"):
    """
    https://zhuanlan.zhihu.com/p/382638682
    :param model:
    :param input_size:
    :param sparsity:
    :param output:
    :param device:
    :return:
    """
    from ptflops.flops_counter import get_model_complexity_info

    if not os.path.exists(output):
        os.makedirs(output)
    config = [{
        'sparsity': sparsity,
        'op_types': ['Conv2d']
        # 'op_types': ['BatchNorm2d']
    }]
    dummy_input = torch.randn(input_size).to(device)
    #
    origin_path = os.path.join(output, 'origin.pth')
    onnx_path = os.path.join(output, 'origin.onnx')
    torch.save(model.state_dict(), origin_path)
    torch.onnx.export(model,
                      dummy_input,
                      onnx_path,
                      do_constant_folding=True,
                      verbose=False,
                      export_params=True,
                      opset_version=11,
                      input_names=['input'],
                      output_names=['output'])

    tmp_model = copy.deepcopy(model).to(device)
    out_tensor = tmp_model(dummy_input)
    flops, params, _ = count_flops_params(tmp_model, dummy_input, verbose=True)
    print(f"Model FLOPs {flops / 1e6:.2f}M, Params {params / 1e6:.2f}M")

    pruner = pruning.FPGMPruner(tmp_model, config)
    pruner.compress()
    flops, params, _ = count_flops_params(tmp_model, dummy_input, verbose=True)
    print(f"Model FLOPs {flops / 1e6:.2f}M, Params {params / 1e6:.2f}M")

    pruned_model_path = os.path.join(output, 'slim_pruned.pth')
    pruned_model_mask = os.path.join(output, 'slim_pruned_mask.pth')
    pruned_model_onnx = os.path.join(output, 'slim_pruned.onnx')

    pruner.export_model(model_path=pruned_model_path, mask_path=pruned_model_mask,
                        onnx_path=pruned_model_onnx, input_shape=dummy_input.shape,
                        device=device,
                        opset_version=11)
    tmp_model = copy.deepcopy(model).to(device)
    print('model pruned done.')
    # apply_compression_results
    apply_compression_results(tmp_model, masks_file=pruned_model_mask)
    print('apply_compression_results')
    out_tensor = tmp_model(dummy_input)
    # Speedup
    m_speedup = ModelSpeedup(tmp_model, dummy_input, masks_file=pruned_model_mask)
    m_speedup.speedup_model()
    out_tensor = tmp_model(dummy_input)
    print('speedup_model ')
    # # print(model)
    slim_speedup_path = os.path.join(output, 'slim_speedup_model.pth')
    slim_speedup_onnx = os.path.join(output, 'slim_speedup_model.onnx')
    torch.save(tmp_model.state_dict(), slim_speedup_path)
    torch.onnx.export(tmp_model, dummy_input, slim_speedup_onnx, verbose=False, opset_version=11)
    print('pruned model exported.')
    flops, params, _ = count_flops_params(tmp_model, dummy_input, verbose=True)
    print(f"Model FLOPs {flops / 1e6:.2f}M, Params {params / 1e6:.2f}M")


class SimpleModel(nn.Module):
    def __init__(self):
        super(SimpleModel, self).__init__()
        self.conv1 = nn.Conv2d(3, 32, 3)
        self.conv2 = nn.Conv2d(32, 64, 3)
        self.conv3 = nn.Conv2d(64, 128, 3)
        self.fc = nn.Linear(128, 256)
        self.classifier = nn.Linear(256, 10)

    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        x = F.relu(self.conv3(x))
        x = nn.functional.adaptive_avg_pool2d(x, 1).reshape(x.shape[0], -1)
        x = self.fc(x)
        x = self.classifier(x)
        return x


if __name__ == "__main__":
    from torchvision.models.resnet import resnet50, resnet18
    from torchvision.models.squeezenet import SqueezeNet
    from torchvision.models.mobilenet import MobileNetV2
    from segment.models import build_model
    from libs.performance import performance

    device = "cuda:0"
    batch_size = 2
    width_mult = 1.0
    num_classes = 20
    input_size = [1, 3, 224, 224]
    net_type = 'modnet_v2'
    model = resnet18(pretrained=True)
    # model = SimpleModel()
    # model = MobileNetV2()
    # model = build_model.get_models(net_type, input_size, num_classes, width_mult=width_mult, pretrained=False)
    model.eval()
    inputs = torch.randn(input_size)
    model = model.to((device))
    inputs = inputs.to((device))
    out = model(inputs)
    performance.model_performance(model, inputs)
    prune_model = copy.deepcopy(model)
    prune_model = model_pruning(prune_model, input_size=input_size, sparsity=0.9, dependency_aware=True, device=device)
    performance.model_performance(model, inputs)
    performance.model_performance(prune_model, inputs)






import numpy as np
from sklearn.metrics import roc_curve, accuracy_score
from sklearn import metrics
from . import roc


class AverageMeter(object):
    """计算并存储参数当前值或平均值
    Computes and stores the average and current value
    -------------------------------------------------
    batch_time = AverageMeter()
    即 self = batch_time
    则 batch_time 具有__init__，reset，update三个属性，
    直接使用batch_time.update()调用
    功能为：batch_time.update(time.time() - end)
       仅一个参数，则直接保存参数值
    对应定义：def update(self, val, n=1)
    losses.update(loss.data[0], input.size(0))
    top1.update(prec1[0], input.size(0))
    top5.update(prec5[0], input.size(0))
    这些有两个参数则求参数val的均值，保存在avg中
    """

    def __init__(self):
        self.reset()  # __init__():reset parameters

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        if n == 0:
            return
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


class MultiAverageMeter(object):
    def __init__(self, keys: list):
        self.keys = keys
        self.avg_meters = {}
        for k in keys: self.avg_meters[k] = AverageMeter()

    def reset(self, keys=None):
        if isinstance(keys, str):
            self.avg_meters[keys].reset()
        elif isinstance(keys, list):
            for k in keys: self.avg_meters[k].reset()
        else:
            for k in self.avg_meters.keys(): self.avg_meters[k].reset()

    def update(self, items: dict, n=1):
        for k, v in items.items(): self.avg_meters[k].update(v, n)


def accuracy(pred, target, topk=(1,)):
    """
    Computes the precision@k for the specified values of k
    :param pred: pred labels,     torch.Size([batch_size, num_classes])
    :param target: target labels, torch.Size([batch_size])
    :param topk: Top K
    :return:
    """
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = pred.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))

    return res


class ROCMeter(object):
    """Compute TPR with fixed FPR"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.target = np.ones(0)
        self.output = np.ones(0)

    def update(self, output, target):
        """
        Usage:
            test_metrics.update(target.cpu().detach().numpy(), output.cpu().detach().numpy())
            tpr001 = self.test_metrics.get_tpr(0.01)
            metrics_i_string = 'TPR@FPR=10-2: {:.4f}\t'.format(tpr001)
        :param output: <class 'tuple'>: (batch_size, num_classes)
        :param target: <class 'tuple'>: (batch_size,)
        :return:
        """
        # If we use cross-entropy
        if len(output.shape) > 1 and output.shape[1] > 1:
            output = output[:, 1]
        elif len(output.shape) > 1 and output.shape[1] == 1:
            output = output[:, 0]
        self.target = np.hstack([self.target, target])
        self.output = np.hstack([self.output, output])

    def get_tpr(self, fixed_fpr=0.01):
        """
        get TPR
        :param fixed_fpr:<float>
        :return:
        """
        fpr, tpr, thr = roc_curve(self.target, self.output, pos_label=1)
        tpr_filtered = tpr[fpr <= fixed_fpr]
        if len(tpr_filtered) == 0:
            return 0.0
        return tpr_filtered[-1]

    def get_tpr_auc(self, fixed_fpr=0.01, plot_roc=True):
        """
        get TPR and AUC
        :param fixed_fpr:<float>
        :return:
        """
        fpr, tpr, thr = roc_curve(self.target, self.output, pos_label=1)
        tpr_filtered = tpr[fpr <= fixed_fpr]
        # 计算auc的值
        roc_auc = metrics.auc(fpr, tpr)

        # 绘制ROC曲线
        if plot_roc:
            fpr_list = [fpr]
            tpr_list = [tpr]
            roc_auc_list = [roc_auc]
            line_names = [""]
            roc.plot_roc_curve(fpr_list, tpr_list, roc_auc_list, line_names=line_names)

        if len(tpr_filtered) == 0:
            return 0.0, 0.0
        return tpr_filtered[-1], roc_auc

    def get_accuracy(self, thr=0.5):
        """
        get Acc
        :param thr:
        :return:
        """
        acc = accuracy_score(self.target,
                             self.output >= thr)
        return acc

    def get_top_hard_examples(self, top_n=10):
        diff_arr = np.abs(self.target - self.output)
        hard_indexes = np.argsort(diff_arr)[::-1]
        hard_indexes = hard_indexes[:top_n]
        return hard_indexes, self.target[hard_indexes], self.output[hard_indexes]



# -*- coding:utf-8 -*-
import os
import argparse
import numbers
import easydict
import yaml
import copy
import json
from . import file_utils


def parser_work_space(work_dir, flags: list = [], time=True):
    """生成工程空间
    flag = [cfg.net_type, cfg.width_mult, cfg.input_size[0], cfg.input_size[1],
            cfg.loss_type, cfg.optim_type, flag, file_utils.get_time()]
    """
    if isinstance(flags, str):
        flags = [flags]
    if time:
        flags += [file_utils.get_time()]
    name = [str(n) for n in flags if n]
    name = "_".join(name)
    work_dir = os.path.join(work_dir, name)
    return work_dir


def parser_config(args: argparse.Namespace, cfg_updata: bool = True):
    """
    解析并合并配置参数：(1)命令行argparse (2)使用*.yaml配置文件
    :param args: 命令行参数
    :param cfg_updata:True: 合并配置参数时，相同参数由*.yaml文件参数决定
                     False: 合并配置参数时，相同参数由命令行argparse参数决定
    :return:
    """
    if "config_file" in args and args.config_file:
        cfg = load_config(args.config_file)
        if cfg_updata:
            cfg = update_dict(args.__dict__, cfg)
        else:
            cfg = update_dict(cfg, args.__dict__)
        cfg["config_file"] = args.config_file
    else:
        cfg = args.__dict__
        cfg['config_file'] = save_config(cfg, 'args_config.yaml')
    print_dict(cfg)
    cfg = easydict.EasyDict(cfg)
    return cfg


def easy2dict(config: easydict.EasyDict):
    """
    :param config: EasyDict参数
    """
    # fix a Bug: cfg = dict(config) 仅仅转换第一层easydict
    cfg = json.loads(json.dumps(config))
    return cfg


def parser_config_file(config: easydict.EasyDict, config_file: str, cfg_updata: bool = True):
    """
    解析并合并配置参数
    :param config: EasyDict参数
    :param cfg_updata:True: 合并配置参数时，相同参数由config参数决定
                     False: 合并配置参数时，相同参数由config_file中的参数决定
    :return:
    """
    cfg = load_config(config_file)
    if cfg_updata:
        cfg = update_dict(cfg, easy2dict(config))
    else:
        cfg = update_dict(easy2dict(config), cfg)
    print_dict(cfg)
    cfg = easydict.EasyDict(cfg)
    return cfg


def parser_config_name(config: dict or str, key="", join="_"):
    """
    :param config: config
    :param key:
    :param join:拼接方式，为空返回list
    :return: List or String
    """
    if isinstance(config, dict) or isinstance(config, easydict.EasyDict):
        cfg = copy.deepcopy(config)
        if key:
            cfg = config[key].copy()
        if isinstance(cfg, dict):
            cfg = list(cfg.keys())
        if not isinstance(cfg, list):
            cfg = [cfg]
        assert isinstance(cfg, list)
        cfg = [str(c) for c in cfg if c]
        if join:
            cfg = "{}".format(join).join(cfg)
    else:
        cfg = config
    return cfg


def update_dict(cfg1: dict, cfg2: dict):
    """相同参数由后面的cfg2决定"""
    # 不同参数进行合并,相同参数由cfg2替换(如果参数是dict类型，会被cfg2整体替换)
    cfg = dict(cfg1, **cfg2)
    # 查找相同参数，cfg1中被cfg2替换，但没有更新的参数
    for k1, v1 in cfg1.items():
        if isinstance(v1, dict):
            cfg[k1] = update_dict(cfg1[k1], cfg2[k1]) if k1 in cfg2 else v1
    return cfg


class Dict2Obj:
    '''
    dict转类对象
    '''

    def __init__(self, args):
        self.__dict__.update(args)


def load_config(config_file='config.yaml'):
    """
    读取配置文件，并返回一个python dict 对象
    :param config_file: 配置文件路径
    :return: python dict 对象
    """
    with open(config_file, 'r', encoding="UTF-8") as stream:
        try:
            config = yaml.load(stream, Loader=yaml.FullLoader)
            # config = Dict2Obj(config)
        except yaml.YAMLError as e:
            print(e)
            return None
    return config


def save_config(cfg: dict, config_file='config.yaml'):
    """保存yaml文件"""
    if isinstance(cfg, easydict.EasyDict) or isinstance(cfg, argparse.Namespace):
        cfg = cfg.__dict__
    fw = open(config_file, 'w', encoding='utf-8')
    yaml.dump(cfg, fw)
    return config_file


def print_dict(dict_data, save_path=None):
    list_config = []
    print("=" * 60)
    for key in dict_data:
        info = "{}: {}".format(key, dict_data[key])
        print(info)
        list_config.append(info)
    if save_path is not None:
        with open(save_path, "w") as f:
            for info in list_config:
                f.writelines(info + "\n")
    print("=" * 60)


if __name__ == '__main__':
    data = None
    # config_file = "config.yaml"
    # save_config(data, config_file)
    cfg1 = {"A": 0, "B": {"X": 0, "Y": 0}, "C": {"ADCD"}}
    cfg2 = {"A": 1, "B": {"X": 1, "Z": {"A": 1}}, "D": 1}
    cfg = update_dict(cfg1, cfg2)
    cfg = easydict.EasyDict(cfg)
    print(cfg)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-07-28 09:09:32
"""

import torch
import torch.utils.data as torch_utils
from tqdm import tqdm
from ..engine import base


class Engine(base.Base):
    def __init__(self, cfg):
        self.cfg = cfg
        self.num_epochs = 0
        self.start_epoch = 0
        self.callbacks = []
        self.model = None
        self.device = "cuda"
        self.optimizer = None
        self.criterion = None
        self.train_loader = None
        self.test_loader = None
        self.progress = True  # 是否显示进度条
        super(Engine, self).__init__()

    def to(self, device):
        self.device = device

    def set_model(self, model):
        self.model = model
        for callback in self.callbacks:
            callback.set_model(self.model)

    def on_batch_begin(self, batch, logs: dict = {}):
        for callback in self.callbacks:
            callback.on_batch_begin(batch, logs)

    def on_batch_end(self, batch, logs: dict = {}):
        for callback in self.callbacks:
            callback.on_batch_end(batch, logs)

    def on_epoch_begin(self, epoch, logs: dict = {}):
        for callback in self.callbacks:
            callback.on_epoch_begin(epoch, logs)

    def on_epoch_end(self, epoch, logs: dict = {}):
        for callback in self.callbacks:
            callback.on_epoch_end(epoch, logs)

    def on_train_begin(self, logs: dict = {}):
        for callback in self.callbacks:
            callback.on_train_begin(logs)

    def on_train_end(self, logs: dict = {}):
        for callback in self.callbacks:
            callback.on_train_end(logs)

    def on_test_begin(self, logs: dict = {}):
        for callback in self.callbacks:
            callback.on_test_begin(logs)

    def on_test_end(self, logs: dict = {}):
        for callback in self.callbacks:
            callback.on_test_end(logs)

    def on_train_summary(self, inputs, outputs, losses, epoch, batch, logs: dict = {}):
        for callback in self.callbacks:
            callback.on_train_summary(inputs, outputs, losses, epoch, batch, logs)

    def on_test_summary(self, inputs, outputs, losses, epoch, batch, logs: dict = {}):
        for callback in self.callbacks:
            callback.on_test_summary(inputs, outputs, losses, epoch, batch, logs)

    def _device(self, inputs):
        if isinstance(inputs, torch.Tensor):
            inputs = inputs.to(self.device)
        elif isinstance(inputs, tuple):
            inputs = (self._device(v) for v in inputs)
        elif isinstance(inputs, list):
            inputs = [self._device(v) for v in inputs]
        elif isinstance(inputs, dict):
            inputs = {k: self._device(v) for k, v in inputs.items()}
        return inputs

    def _losses(self, losses):
        if isinstance(losses, dict):
            l = sum(losses.values())
        elif isinstance(losses, tuple) or isinstance(losses, list):
            l = sum(losses)
        else:
            l = losses
        return l

    def run(self, logs: dict = {}):
        self.model.to(self.device)
        self.on_train_begin(logs)
        for epoch in range(self.num_epochs):
            self.epoch = epoch
            self.set_model(self.model)
            self.on_epoch_begin(epoch, logs)
            if self.train_loader:
                self.run_train_epoch(self.train_loader, logs)
            self.on_test_begin(logs)
            if self.test_loader:
                self.run_test_epoch(self.test_loader, logs)
            self.on_test_end(logs)
            self.on_epoch_end(epoch, logs)
        self.on_train_end(logs)

    def run_train_epoch(self, dataset: torch_utils.DataLoader, logs: dict = {}):
        num_steps = len(dataset)
        batch = 0
        self.model.train()  # set to training mode
        with tqdm(total=num_steps, desc="Train Epoch #{}".format(self.epoch), disable=not self.progress) as t:
            for inputs in dataset:
                self.on_batch_begin(batch, logs)
                inputs = self._device(inputs)
                outputs, losses = self.run_step(inputs)
                # compute gradient and do optimizer step
                self.optimizer.zero_grad()
                self._losses(losses).backward()
                self.optimizer.step()
                logs['lr'] = self.optimizer.param_groups[0]["lr"]
                self.on_train_summary(inputs, outputs, losses, self.epoch, batch, logs)
                self.on_batch_end(batch, logs)
                batch += 1
                t.update(1)

    def run_test_epoch(self, dataset: torch_utils.DataLoader, logs: dict = {}):
        num_steps = len(dataset)
        batch = 0
        self.model.eval()  # set to evaluates mode
        with tqdm(total=num_steps, desc="Test Epoch #{}".format(self.epoch), disable=not self.progress) as t:
            for inputs in dataset:
                with torch.no_grad():
                    inputs = self._device(inputs)
                    outputs, losses = self.run_step(inputs)
                    self.on_test_summary(inputs, outputs, losses, self.epoch, batch, logs)
                batch += 1
                t.update(1)

    def run_step(self, inputs):
        data, targets = inputs
        outputs = self.model(data)
        losses = self.criterion(outputs, targets)
        return outputs, losses



# -*-coding: utf-8 -*-
"""
    @Author : PKing
    @E-mail : 390737991@qq.com
    @Date   : 2022-12-23 14:45:13
    @Brief  :
"""
import os
import cv2
import numpy as np
import pycuda.driver as cuda
import tensorrt as trt
from typing import Tuple, List

cuda.init()
device_context = cuda.Device(0).make_context()
TRT_LOGGER = trt.Logger(trt.Logger.ERROR)  # TRT_LOGGER = trt.Logger(trt.Logger.WARNING)


class TRTEngine(object):
    """TensorRT引擎"""

    def __init__(self, model_file, input_shape):
        """
        https://www.jianshu.com/p/36ff0e224112
        :param model_file: 模型文件，可以是*.trt或者*.onnx文件
        :param input_shape:输入维度(B, C, H, W)
        """
        print("tensorrt:{}".format(trt.__version__))  # 8.4.1.5
        assert len(input_shape) == 4
        B, C, H, W = input_shape
        self.input_size = (W, H)
        self.batch_size = B
        self.input_shape = input_shape
        # Load a serialized engine into memory
        # self.engine = self.load_engine(model_file)  # 加载序列化的cuda引擎
        self.engine = self.build_engine_onnx(model_file=model_file, input_shape=self.input_shape)
        # Create context, this can be re-used  创建 执行环境
        self.context = self.engine.create_execution_context()
        # Profile 0 (first profile) is used by default  context可以设置多个profile， 这里选择第一个，也是默认的profile，其中规定了输入尺寸的变化区间
        self.context.active_optimization_profile = 0
        print("Active Optimization Profile: {}".format(self.context.active_optimization_profile))
        # These binding_idxs can change if either the context or the
        # active_optimization_profile are changed  获得输入输出变量名对应profile的idx
        self.input_binding_idxs, self.output_binding_idxs = self.get_binding_idxs(self.engine,
                                                                                  self.context.active_optimization_profile)
        print("\tinput_shape                    : {}".format(self.input_shape))
        # 获得输入变量的变量名
        self.input_names = [self.engine.get_binding_name(binding_idx) for binding_idx in self.input_binding_idxs]
        print("Input Metadata")
        print("\tNumber of Inputs: {}".format(len(self.input_binding_idxs)))
        print("\tInput names     : {}".format(self.input_names))
        print("\tInput Bindings for Profile {}: {}".format(self.context.active_optimization_profile,
                                                           self.input_binding_idxs))
        print("\tInput names: {}".format(self.input_names))
        # 获得输出变量的变量名
        print("Output Metadata")
        self.output_names = [self.engine.get_binding_name(binding_idx) for binding_idx in self.output_binding_idxs]
        print("\tNumber of Outputs: {}".format(len(self.output_binding_idxs)))
        print("\tOutput names     : {}".format(self.output_names))
        print("\tOutput Bindings for Profile {}: {}\n".format(self.context.active_optimization_profile,
                                                              self.output_binding_idxs), flush=True)

    @staticmethod
    def build_engine_onnx(model_file: str, input_shape, fp16=True):
        """
        :param model_file: onnx file or trt file
        :param batch_size:
        :param fp16_mode:
        :param max_workspace:
        :return:
        """
        if model_file.endswith("trt") and os.path.exists(model_file):
            trt_file = model_file
            print('Reload engine file: {}'.format(trt_file))
            engine = TRTEngine.load_engine(trt_file)
            return engine
        elif model_file.endswith("onnx"):
            onnx_file = model_file
            trt_file = model_file[:-len("onnx")] + "trt"
        else:
            raise Exception("model_file:{}".format(model_file))
        print("Completed parsing of ONNX file")
        print("convert ONNX file to TRT Engine")
        print("load ONNX  file:{}".format(onnx_file))
        engine = onnx2trt(onnx_file, trt_file=trt_file, input_shape=input_shape, fp16=fp16)
        return engine

    def __call__(self, input_tensor):
        return self.inference(input_tensor)

    def inference(self, image_tensor):
        """
        image_tensor = image.transpose(2, 0, 1)
        image_tensor = image_tensor[np.newaxis, :]
        :param image_tensor:
        :return:
        """
        input_tensor = [image_tensor]
        outputs = self.forward(input_tensor)
        return outputs

    def forward(self, input_tensor):
        device_context.push()
        # Allocate device memory for inputs. This can be easily re-used if the
        # input shapes don't change  为输入变量赋予host空间，该空间可复用
        device_inputs = [cuda.mem_alloc(h_input.nbytes) for h_input in input_tensor]
        # Copy host inputs to device, this needs to be done for each new input， 由host拷贝到device
        for h_input, d_input in zip(input_tensor, device_inputs):
            cuda.memcpy_htod(d_input, h_input)
        # This needs to be called everytime your input shapes change
        # If your inputs are always the same shape (same batch size, etc.),
        # then you will only need to call this once 重新指定网络输入输出的大小。
        host_outputs, device_outputs = self.setup_binding_shapes(self.engine, self.context, input_tensor,
                                                                 self.input_binding_idxs,
                                                                 self.output_binding_idxs)  # 返回的是输出的idx和device buffer

        # print("\tOutput shapes: {}".format([out.shape for out in host_outputs]))
        # Bindings are a list of device pointers for inputs and outputs
        bindings = device_inputs + device_outputs  # list的合并
        # Inference
        self.context.execute_v2(bindings)
        # Copy outputs back to host to view results 将输出由gpu拷贝到cpu。
        for h_output, d_output in zip(host_outputs, device_outputs):
            cuda.memcpy_dtoh(h_output, d_output)
        # 释放显存
        for b in bindings: b.free()
        device_context.pop()
        return host_outputs

    @staticmethod
    def load_engine(filename: str):
        # Load serialized engine file into memory 加载序列化的cuda引擎并进行反序列化
        with open(filename, "rb") as f, trt.Runtime(TRT_LOGGER) as runtime:
            return runtime.deserialize_cuda_engine(f.read())

    @staticmethod
    def get_binding_idxs(engine: trt.ICudaEngine, profile_index: int):
        # Calculate start/end binding indices for current context's profile
        num_bindings_per_profile = engine.num_bindings // engine.num_optimization_profiles
        start_binding = profile_index * num_bindings_per_profile
        end_binding = start_binding + num_bindings_per_profile
        print("Engine/Binding Metadata")
        print("\tNumber of optimization profiles: {}".format(engine.num_optimization_profiles))
        print("\tNumber of bindings per profile : {}".format(num_bindings_per_profile))
        print("\tFirst binding for profile {}   : {}".format(profile_index, start_binding))
        print("\tLast binding for profile {}    : {}".format(profile_index, end_binding - 1))

        # Separate input and output binding indices for convenience
        input_binding_idxs = []
        output_binding_idxs = []
        for binding_index in range(start_binding, end_binding):
            if engine.binding_is_input(binding_index):
                input_binding_idxs.append(binding_index)
            else:
                output_binding_idxs.append(binding_index)

        return input_binding_idxs, output_binding_idxs

    @staticmethod
    def setup_binding_shapes(
            engine: trt.ICudaEngine,
            context: trt.IExecutionContext,
            host_inputs: List[np.ndarray],
            input_binding_idxs: List[int],
            output_binding_idxs: List[int],
    ):
        # 指定输入的shape，同时根据输入的shape指定输出的shape，并未输出赋予cuda空间
        # Explicitly set the dynamic input shapes, so the dynamic output
        # shapes can be computed internally
        for host_input, binding_index in zip(host_inputs, input_binding_idxs):
            context.set_binding_shape(binding_index, host_input.shape)
        # assert context.all_binding_shapes_specified
        host_outputs = []
        device_outputs = []
        for binding_index in output_binding_idxs:
            output_shape = context.get_binding_shape(binding_index)
            # Allocate buffers to hold output results after copying back to host
            buffer = np.empty(output_shape, dtype=np.float32)
            host_outputs.append(buffer)
            # Allocate output buffers on device
            device_outputs.append(cuda.mem_alloc(buffer.nbytes))
        return host_outputs, device_outputs


def onnx2trt(onnx_file, trt_file=None, input_shape=(1, 3, 160, 160), max_batch_size=8, fp16=True):
    """
    :param onnx_file: ONNX模型文件
    :param trt_file: 输出TRT模型文件，默认为onnx_file同目录
    :param input_shape: 模型输入维度
    :param fp16: 是否进行半精度量化
    :return:
    """

    def GiB(val):
        return val * 1 << 30

    if not trt_file: trt_file = onnx_file[:-len("onnx")] + "trt"
    batch_size, C, H, W = input_shape
    explicit_batch = 1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
    with trt.Builder(TRT_LOGGER) as builder, \
            builder.create_network(explicit_batch) as network, \
            trt.OnnxParser(network, TRT_LOGGER) as parser:
        builder.max_batch_size = max_batch_size  # trt推理时最大支持的batchsize
        config = builder.create_builder_config()
        config.max_workspace_size = GiB(4)
        if fp16: config.set_flag(trt.BuilderFlag.FP16)
        # if int8:config.set_flag(trt.BuilderFlag.INT8)
        print('Loading ONNX file from path {}...'.format(onnx_file))
        with open(onnx_file, 'rb') as model:
            print('Beginning ONNX file parsing')
            parser.parse(model.read())
        print('Completed parsing of ONNX file')
        print('Building an engine from file {}; this may take a while...'.format(onnx_file))
        profile = builder.create_optimization_profile()  # 动态输入时候需要 分别为最小输入、常规输入、最大输入
        # 有几个输入就要写几个profile.set_shape 名字和转onnx的时候要对应
        # tensorrt6以后的版本是支持动态输入的，需要给每个动态输入绑定一个profile，用于指定最小值，常规值和最大值，如果超出这个范围会报异常。
        profile.set_shape("input", (1, C, H, W), (batch_size, C, H, W), (batch_size * 2, C, H, W))
        config.add_optimization_profile(profile)
        engine = builder.build_engine(network, config)
    print("Completed creating Engine")
    # 保存engine文件
    with open(trt_file, "wb") as f:
        f.write(engine.serialize())
    print("save TRT file:{}".format(trt_file))
    return engine


if __name__ == "__main__":
    trt_file = "/home/dm/nasdata/release/handwriting/daip-calligraphy-hard/calligraphy-hard-recognizer/app/infercore/resource/resnet18_160_160_c3594_7.onnx"
    # trt_file = "/home/dm/nasdata/release/handwriting/daip-calligraphy-hard/calligraphy-hard-recognizer/app/infercore/resource/resnet18_160_160_c3594_7.trt"
    input_shape = (16, 3, 160, 160)  # (B,C,H,W)
    trt = TRTEngine(trt_file, input_shape=input_shape)
    image = np.ones(shape=input_shape)
    input_tensor = np.asarray(image, dtype=np.float32)
    outputs, = trt(input_tensor)
    print("input_tensor:{}".format(input_tensor.shape))
    print("output      :{}={}".format(outputs.shape, outputs))



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-07-28 14:57:10
"""
import sys
import os
import torch
from ..callbacks.callbacks import Callback
from ..utils import summary
from ..utils import log, file_utils


class ModelCheckpoint(Callback):
    def __init__(self,
                 model: torch.nn.Module,
                 optimizer: torch.optim,
                 moder_dir: str,
                 epochs: int,
                 start_save: int = -1,
                 indicator: str = "",
                 inverse: bool = False,
                 logger=None,
                 ):
        """
        保存模型回调函数
        :param model: 模型
        :param optimizer:优化器
        :param moder_dir:保存训练模型的目录
        :param epochs: 训练的epochs数
        :param start_save: epoch >= start_save开始保存，如果为-1，则保存最后10个epoch模型
        :param indicator:需要关注的指标，以便保存最优模型，需要根据Metrics定义的指标对应，
                         如分类模型中indicator="acc"；如果关注losss,则indicator="loss"
                         如果不需要关注，则设置为空
                         PS: 优先关注"test"的指标
        :param inverse: 关注的指标(indicator)值越大越好，则inverse=True；值越小越好，则inverse=False
        :param logger: Log实例对象
        """
        super().__init__()
        self.model = model
        self.optimizer = optimizer
        self.moder_dir = moder_dir
        self.epochs = epochs
        self.start_save = start_save if start_save else -1
        file_utils.create_dir(self.moder_dir)
        self.logger = log.get_logger() if logger is None else logger
        self.main_process = True
        self.indicator = indicator
        self.indicator_val = 0
        self.inverse = inverse
        if self.inverse:
            self.indicator_val = sys.maxsize  # 评价指标

    def get_indicators_values(self, indicator, logs):
        r = None
        for k, v in logs.items():
            if k == indicator:
                return v
            elif isinstance(v, dict):
                r = self.get_indicators_values(indicator, v)
        return r

    def _sum(self, value):
        if isinstance(value, dict):
            # 去除total的数据，避免重复计算求和
            l = {k: v for k, v in value.items() if "total_" not in k}
            l = sum(l.values())
        elif isinstance(value, tuple) or isinstance(value, list):
            l = sum(value)
        else:
            l = value
        return l

    def on_epoch_end(self, epoch, logs: dict = {}):
        # fix a bug:
        if not self.is_main_process: return
        if "test" in logs:
            value = self.get_indicators_values(self.indicator, logs["test"])
        else:
            value = self.get_indicators_values(self.indicator, logs)
        value = self._sum(value)
        self.save_model(self.moder_dir, value, epoch, start_save=self.start_save)
        if value:
            self.save_best_model(self.moder_dir, value, epoch, self.inverse)

    def save_model(self, model_root, value, epoch, start_save=0):
        """保存模型"""
        model = self.model
        optimizer = self.optimizer
        if value:
            name = "model_{:0=3d}_{:.4f}.pth".format(epoch, value)
        else:
            name = "model_{:0=3d}.pth".format(epoch)
        # 保存当前优化器和模型
        torch.save({"epoch": epoch,
                    "model": model.module.state_dict(),
                    "optimizer": optimizer.state_dict()},
                   os.path.join(model_root, "model_optimizer.pth"))
        # 保存最新的模型
        model_file = os.path.join(model_root, "latest_{}".format(name))
        file_utils.remove_prefix_files(model_root, "latest_*")
        torch.save(model.module.state_dict(), model_file)
        # 保存最后epoch模型
        start_save = start_save if start_save >= 0 else self.epochs - 10
        if epoch >= start_save:
            model_file = os.path.join(model_root, name)
            torch.save(model.module.state_dict(), model_file)
        self.logger.info("save model:{}".format(model_file))

    def save_best_model(self, model_root, value, epoch, inverse=False):
        """保存关注的指标(indicator)最优的模型"""
        model = self.model
        optimizer = self.optimizer
        if value > self.indicator_val and inverse:
            # indicator指标，值越小，性能越好
            return
        elif value < self.indicator_val and (not inverse):
            # indicator指标，值越大，性能越好
            return
        self.indicator_val = value
        model_file = os.path.join(model_root, "best_model_{:0=3d}_{:.4f}.pth".format(epoch, value))
        file_utils.remove_prefix_files(model_root, "best_model_*")
        torch.save(model.module.state_dict(), model_file)
        self.logger.info("save best_model:{}".format(model_file))

    def resume_model(self, model, optimizer, resume, strict=True):
        """
        optimizer_pth = os.path.join(self.model_root, "optimizer_{}.pth".format(self.net_type))
        resume or finetune model
        torch.save({"epoch": epoch,
                    "model": model.state_dict(),
                    "optimizer": optimizer.state_dict()}, optimizer_pth)
        :return:
        """
        start_epoch = 0
        if os.path.isdir(resume):
            optimizer_pth = os.path.join(self.moder_dir, "model_optimizer.pth")
            model, optimizer, start_epoch = self._resume_model(optimizer_pth, model, optimizer)
            self.logger.info("resume_model:{},start_epoch:{}".format(optimizer_pth, start_epoch))
        elif os.path.isfile(resume):
            state_dict = torch.load(resume, map_location="cpu")
            if "state_dict" in state_dict:
                state_dict = state_dict["state_dict"]
            model.load_state_dict(state_dict, strict=strict)
            self.logger.info("pretrain_model:{},start_epoch:{},strict:{}".format(resume, start_epoch, strict))
        else:
            self.logger.info("no resume_model:{},start_epoch:{}".format(resume, start_epoch))
        return model, optimizer, start_epoch

    def _resume_model(self, optimizer_pth, model, optimizer):
        """
        optimizer_pth = os.path.join(self.model_root, "optimizer_{}.pth".format(self.net_type))
        resume or finetune model
        torch.save({"epoch": epoch,
                    "model": model.state_dict(),
                    "optimizer": optimizer.state_dict()}, optimizer_pth)
        :return:
        """
        state_dict = torch.load(optimizer_pth, map_location="cpu")
        model.load_state_dict(state_dict["model"])
        # load optimizer parameter
        optimizer.load_state_dict(state_dict["optimizer"])
        for state in optimizer.state.values():
            for k, v in state.items():
                if torch.is_tensor(v):
                    state[k] = v
        start_epoch = state_dict["epoch"]
        return model, optimizer, start_epoch



# -*- coding: utf-8 -*-
"""
# --------------------------------------------------------
# @Author : panjq
# @Date   : 2019-9-20 13:18:34
# --------------------------------------------------------
"""
import os
import datetime
import logging
import threading
import re
import time
from logging.handlers import TimedRotatingFileHandler


# from memory_profiler import profile


def singleton(cls):
    _instance_lock = threading.Lock()
    instances = {}

    def _singleton(*args, **kwargs):
        with _instance_lock:
            if cls not in instances:
                instances[cls] = cls(*args, **kwargs)
            return instances[cls]

    return _singleton


@singleton  # 使用singleton，会出现loger的level失效的问题
class CustomLogger(logging.Logger):
    def __init__(self, name="LOG", level="debug"):
        """
        Initialize the logger with a name and an optional level.
        Args:
            name:
            level: debug,info,warning,critical,fatal
        """
        super().__init__(name)
        # super(CustomLogger, self).__init__(name)
        self.setLevel(level=level)

    @staticmethod
    def levels(level):
        if level == 'debug':
            return logging.DEBUG
        if level == 'info':
            return logging.INFO
        if level == 'warning':
            return logging.WARN
        if level == 'critical':
            return logging.CRITICAL
        if level == 'fatal':
            return logging.FATAL
        return logging.DEBUG

    def setLevel(self, level):
        """
        Args:
            level: debug,info,warning,critical,fatal
        Returns:
        """
        level = self.levels(level)
        super().setLevel(level)

    @staticmethod
    def set_format(handler, format):
        # handler.suffix = "%Y%m%d"
        if format:
            logFormatter = logging.Formatter("%(asctime)s %(filename)s %(funcName)s %(levelname)s: %(message)s",
                                             "%Y-%m-%d %H:%M:%S")
        else:
            logFormatter = logging.Formatter("%(levelname)s: %(message)s")
        handler.setFormatter(logFormatter)

    def show_batch_tensor(self, title, batch_imgs, index=0):
        pass


class FileHandler(TimedRotatingFileHandler):
    def __init__(self, filename, when='h', interval=1, backupCount=0, encoding=None, delay=False, utc=False,
                 atTime=None):
        logging.handlers.BaseRotatingHandler.__init__(self, filename, 'a', encoding, delay)
        self.when = when.upper()
        self.backupCount = backupCount
        self.utc = utc
        self.atTime = atTime
        if self.when == 'S':
            self.interval = 1  # one second
            self.suffix = "%Y-%m-%d_%H-%M-%S"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}(\.\w+)?$"
        elif self.when == 'M':
            self.interval = 60  # one minute
            self.suffix = "%Y-%m-%d_%H-%M"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}-\d{2}(\.\w+)?$"
        elif self.when == 'H':
            self.interval = 60 * 60  # one hour
            self.suffix = "%Y-%m-%d_%H"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}(\.\w+)?$"
        elif self.when == 'D' or self.when == 'MIDNIGHT':
            self.interval = 60 * 60 * 24  # one day
            self.suffix = "%Y-%m-%d"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}(\.\w+)?$"
        elif self.when.startswith('W'):
            self.interval = 60 * 60 * 24 * 7  # one week
            if len(self.when) != 2:
                raise ValueError("You must specify a day for weekly rollover from 0 to 6 (0 is Monday): %s" % self.when)
            if self.when[1] < '0' or self.when[1] > '6':
                raise ValueError("Invalid day specified for weekly rollover: %s" % self.when)
            self.dayOfWeek = int(self.when[1])
            self.suffix = "%Y-%m-%d"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}(\.\w+)?$"
        elif self.when == 'Y':
            self.interval = 60 * 60 * 24 * 365  # one yes
            self.suffix = "%Y-%m-%d"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}(\.\w+)?$"
        else:
            raise ValueError("Invalid rollover interval specified: %s" % self.when)

        self.extMatch = re.compile(self.extMatch, re.ASCII)
        self.interval = self.interval * interval  # multiply by units requested
        # The following line added because the filename passed in could be a
        # path object (see Issue #27493), but self.baseFilename will be a string
        filename = self.baseFilename
        if os.path.exists(filename):
            t = os.stat(filename)[logging.handlers.ST_MTIME]
        else:
            t = int(time.time())
        self.rolloverAt = self.computeRollover(t)


def set_logger(name="LOG", level="debug", logfile=None, format=False, is_main_process=True):
    """
    logger = set_logging(name="LOG", level="debug", logfile="log.txt", format=False)
    url:https://cuiqingcai.com/6080.html
    level级别：debug>info>warning>error>critical
    :param level: 设置log输出级别
    :param logfile: log保存路径，如果为None，则在控制台打印log
    :param is_main_process: 是否是主进程
    :return:
    """
    if not is_main_process:
        level = "fatal"
        logfile = None
    # logger = logging.getLogger(name)
    logger = CustomLogger(name, level=level)
    if logfile and os.path.exists(logfile):
        os.remove(logfile)
    # define a FileHandler write messages to file
    if logfile:
        # filehandler = logging.handlers.RotatingFileHandler(filename="./log.txt")
        # filehandler = TimedRotatingFileHandler(logfile, when="midnight", interval=1)
        filehandler = FileHandler(logfile, when="Y", interval=1)
        logger.set_format(filehandler, format)
        logger.addHandler(filehandler)

    # define a StreamHandler print messages to console
    console = logging.StreamHandler()
    logger.set_format(console, format)
    logger.addHandler(console)
    return logger


def print_args(args):
    logger = get_logger()
    logger.info("---" * 10)
    args = args.__dict__
    for k, v in args.items():
        # print("{}: {}".format(k, v))
        logger.info("{}: {}".format(k, v))
    logger.info("---" * 10)


def get_logger(name="LOG", level="debug"):
    logger = CustomLogger(name)
    # if logger.isEnabledFor(CustomLogger.levels(level)):
    #     logger = CustomLogger(name, level=level)
    return logger


if __name__ == '__main__':
    logger = set_logger(logfile=None, level="debug")
    logger1 = get_logger()
    logger1.info("---" * 20)
    logger1.debug("work_space:{}".format("work_dir"))
    logger1.info("work_space:{}".format("work_dir"))
    logger1.error("work_space:{}".format("work_dir"))
    logger1.fatal("work_space:{}".format("work_dir"))
    # logger1.show_batch_tensor()



# -*-coding: utf-8 -*-
"""
    @Project: python-learning-notes
    @File   : iou.py
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2019-07-10 10:14:56
"""

import numpy as np


def cal_iou(box1, box2):
    """
    computing IoU
    :param box1: (y0, x0, y1, x1), which reflects
            (top, left, bottom, right)
    :param box2: (y0, x0, y1, x1)
    :return: scala value of IoU
    """
    cx1, cy1, cx2, cy2 = box1
    gx1, gy1, gx2, gy2 = box2
    # 计算每个矩形的面积
    S_rec1 = (cx2 - cx1) * (cy2 - cy1)  # C的面积
    S_rec2 = (gx2 - gx1) * (gy2 - gy1)  # G的面积

    # 计算相交矩形
    x1 = max(cx1, gx1)
    y1 = max(cy1, gy1)
    x2 = min(cx2, gx2)
    y2 = min(cy2, gy2)

    w = max(0, x2 - x1)
    h = max(0, y2 - y1)
    area = w * h  # C∩G的面积

    iou = area / (S_rec1 + S_rec2 - area)
    return iou


def cal_iou_list(box1, box2_list):
    iou_list = []
    for box2 in box2_list:
        iou = cal_iou(box1, box2)
        iou_list.append(iou)
    return iou_list


def get_iou_mat(box1_list, box2_list):
    '''
    获得IOU矩阵:iou_mat=(len1,len2)
    :param box1_list: len1
    :param box2_list: len2
    :return:
    '''
    iou_mat = []
    for box1 in box1_list:
        iou_list = cal_iou_list(box1, box2_list)
        iou_mat.append(iou_list)
    iou_mat = np.asarray(iou_mat)
    return iou_mat


def get_max_iou_index(iou_mat, iou_threshold):
    max_index = np.argmax(iou_mat, axis=1)
    max_iou = np.max(iou_mat, axis=1)
    index = max_iou > iou_threshold
    max_iou = max_iou[index]
    max_index = max_index[index]
    return max_iou, max_index


if __name__ == '__main__':
    box1 = [661, 27, 679, 47]
    box2 = [661, 27, 679, 47]
    box2_list = []
    box2_list.append(box2)
    box2_list.append(box2)
    iou = cal_iou_list(box1, box2_list)
    print(iou)

if __name__ == '__main__':
    box1 = [661, 27, 679, 47]
    box2 = [661, 27, 679, 47]
    box2_list = []
    box2_list.append(box2)
    box2_list.append(box2)
    iou = cal_iou_list(box1, box2_list)
    print(iou)



# -*- coding: utf-8 -*-
"""
# --------------------------------------------------------
# @Author : panjq
# @E-mail : pan_jinquan@163.com
# @Date   : 2021-07-26 10:54:14
# --------------------------------------------------------
"""


class Base(object):
    def __init__(self):
        self.model = None

    def set_model(self, model):
        """设置模型"""
        self.model = model

    def on_batch_begin(self, batch, logs: dict = {}):
        """每次迭代开始时回调"""
        pass

    def on_batch_end(self, batch, logs: dict = {}):
        """每次迭代结束时回调"""
        pass

    def on_train_summary(self, inputs, outputs, losses, epoch, step, logs: dict = {}):
        """每次迭代，训练结束时回调"""
        pass

    def on_test_summary(self, inputs, outputs, losses, epoch, step, logs: dict = {}):
        """每次迭代，测试结束时回调"""
        pass

    def on_epoch_begin(self, epoch, logs: dict = {}):
        """每个epoch开始时回调"""
        pass

    def on_epoch_end(self, epoch, logs: dict = {}):
        """每个epoch结束时回调"""
        pass

    def on_train_begin(self, logs: dict = {}):
        """开始训练时回调"""
        pass

    def on_train_end(self, logs: dict = {}):
        """结束训练时回调"""
        pass

    def on_test_begin(self, logs: dict = {}):
        """开始测试时回调"""
        pass

    def on_test_end(self, logs: dict = {}):
        """结束测试时回调"""
        pass



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-08-13 10:08:07
"""
import torch
import torch.optim as optim
from torchvision import models, transforms
from basetrainer.scheduler import build_scheduler
from basetrainer.scheduler.MultiStepLR import MultiStepLR
from basetrainer.scheduler.CosineAnnealingLR import CosineAnnealingLR
from basetrainer.utils import plot_utils


def test_callback_for_build_scheduler(num_epochs, num_steps, optimizer, lr_init, milestones, num_warn_up):
    scheduler = "ExpLR"  # cosine,multi-step
    # scheduler = "LambdaLR"  # cosine,multi-step
    lr_scheduler = build_scheduler.get_scheduler(scheduler,
                                                 optimizer=optimizer,
                                                 lr_init=lr_init,
                                                 num_epochs=num_epochs,
                                                 num_steps=num_steps,
                                                 milestones=milestones,
                                                 num_warn_up=num_warn_up)

    lr_list = []
    for epoch in range(num_epochs):
        lr_scheduler.on_epoch_begin(epoch)
        for step in range(num_steps):
            lr_scheduler.step(epoch, step)
        lr = optimizer.param_groups[0]["lr"]
        lr_list.append(lr)
        print("epoch:{},lr:{}".format(epoch, lr))
    plot_utils.plot_multi_line(x_data_list=[range(num_epochs)], y_data_list=[lr_list])


def test_torch_scheduler(num_epochs, num_steps, optimizer, lr_init, milestones, num_warn_up):
    # 余弦退火学习率,T_0是周期，T_mult就是之后每个周期T_0 = T_0 * T_mult，eta_min最低学习率
    lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0=2, T_mult=2,eta_min=1e-5)
    # num_cycles = 3
    # t_max = num_epochs * 1.0 / (2 * num_cycles - 1)  # 一次学习率周期的迭代次数，即 T_max 个 epoch 之后重新设置学习率。
    # eta_min = 0.000001  # 最小学习率，即在一个周期中，学习率最小会下降到 eta_min，默认值为 0
    # lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, t_max, eta_min=eta_min, last_epoch=-1)

    lr_list = []
    for epoch in range(num_epochs):
        # lr_scheduler.on_epoch_begin(epoch)
        for step in range(num_steps):
            # lr_scheduler.step(epoch, step)
            pass
        lr_scheduler.step()
        lr = optimizer.param_groups[0]["lr"]
        lr_list.append(lr)
        print("epoch:{},lr:{}".format(epoch, lr))
    plot_utils.plot_multi_line(x_data_list=[range(num_epochs)], y_data_list=[lr_list])


if __name__ == "__main__":
    num_epochs = 150
    num_warn_up = 5
    lr_init = 0.1
    num_steps = 10
    milestones = [50, 100, 120]
    model = models.resnet18(pretrained=False)
    optimizer = optim.SGD(model.parameters(), lr=lr_init)
    test_callback_for_build_scheduler(num_epochs, num_steps, optimizer, lr_init, milestones, num_warn_up)
    # test_torch_scheduler(num_epochs, num_steps, optimizer, lr_init, milestones, num_warn_up)



# -*-coding: utf-8 -*-
"""
    @Author : PKing
    @E-mail : 390737991@qq.com
    @Date   : 2022-12-23 11:06:28
    @Brief  : https://blog.csdn.net/weixin_42492254/article/details/125319112
"""

import os
import tensorrt as trt
import sys

print("tensorrt:{}".format(trt.__version__))  # 8.4.3.1

TRT_LOGGER = trt.Logger()


def onnx2trt(onnx_file, trt_file=None, input_shape=(1, 3, 160, 160), max_batch_size=8, fp16=True):
    """
    :param onnx_file: ONNX模型文件
    :param trt_file: 输出TRT模型文件，默认为onnx_file同目录
    :param input_shape: 模型输入维度
    :param fp16: 是否进行半精度量化
    :return:
    """

    def GiB(val):
        return val * 1 << 30

    if not trt_file: trt_file = onnx_file[:-len("onnx")] + "trt"
    batch_size, C, H, W = input_shape
    explicit_batch = 1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
    with trt.Builder(TRT_LOGGER) as builder, \
            builder.create_network(explicit_batch) as network, \
            trt.OnnxParser(network, TRT_LOGGER) as parser:
        builder.max_batch_size = max_batch_size  # trt推理时最大支持的batchsize
        config = builder.create_builder_config()
        config.max_workspace_size = GiB(4)
        if fp16: config.set_flag(trt.BuilderFlag.FP16)
        # if int8:config.set_flag(trt.BuilderFlag.INT8)
        print('Loading ONNX file from path {}...'.format(onnx_file))
        with open(onnx_file, 'rb') as model:
            print('Beginning ONNX file parsing')
            parser.parse(model.read())
        print('Completed parsing of ONNX file')
        print('Building an engine from file {}; this may take a while...'.format(onnx_file))
        profile = builder.create_optimization_profile()  # 动态输入时候需要 分别为最小输入、常规输入、最大输入
        # 有几个输入就要写几个profile.set_shape 名字和转onnx的时候要对应
        # tensorrt6以后的版本是支持动态输入的，需要给每个动态输入绑定一个profile，用于指定最小值，常规值和最大值，如果超出这个范围会报异常。
        profile.set_shape("input", (1, C, H, W), (batch_size, C, H, W), (batch_size * 2, C, H, W))
        config.add_optimization_profile(profile)
        engine = builder.build_engine(network, config)
    print("Completed creating Engine")
    # 保存engine文件
    with open(trt_file, "wb") as f:
        f.write(engine.serialize())
    print("save TRT file:{}".format(trt_file))
    return engine


if __name__ == "__main__":
    onnx_file = '/home/dm/nasdata/release/handwriting/daip-calligraphy-hard/calligraphy-hard-recognizer/app/infercore/resource/resnet18_160_160_c3594_7.onnx'
    onnx2trt(onnx_file, input_shape=(1, 3, 160, 160))



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-12-10 17:39:21
"""
import os
import torch
import torch.nn as nn
import torch_pruning as pruning


def model_pruning(model: nn.Module,
                  input_size=[1, 3, 128, 128],
                  sparsity=0.2,
                  prune_mod="fpgm",
                  prune_shortcut=False,
                  device="cpu",
                  **kwargs):
    """
    使用Slim-pruner进行模型剪枝和压缩
    https://github.com/yanggui19891007/Pytorch-Auto-Slim-Tools
    :param model:  Pytorch模型
    :param input_size: 模型输入的维度[batch-size,channel,H,W]
    :param sparsity: 模型稀疏力度,值越大,模型压缩越大，计算量越小；
    :param device:
    :return:
    """
    input_res = tuple(input_size[1:])
    model.cpu()
    dummy_input = torch.randn(input_size).cpu()
    flops_raw, params_raw = pruning.get_model_complexity_info(model,
                                                              input_res=input_res,
                                                              as_strings=True,
                                                              print_per_layer_stat=False)
    print('-[INFO] before pruning flops:  ' + flops_raw)
    print('-[INFO] before pruning params:  ' + params_raw)
    # 剪枝引擎建立
    slim = pruning.Autoslim(model, inputs=dummy_input, compression_ratio=sparsity)
    if prune_mod == 'fpgm':
        config = {
            'layer_compression_ratio': None,
            'norm_rate': 1.0, 'prune_shortcut': 1,
            'dist_type': 'l1', 'pruning_func': 'fpgm'
        }
    elif prune_mod == 'l1':
        config = {
            'layer_compression_ratio': None,
            'norm_rate': 1.0, 'prune_shortcut': 1,
            'global_pruning': False, 'pruning_func': 'l1'
        }

    slim.base_prunging(config)
    flops_new, params_new = pruning.get_model_complexity_info(model,
                                                              input_res,
                                                              as_strings=True,
                                                              print_per_layer_stat=False)
    print('\n-[INFO] after pruning flops:  ' + flops_new)
    print('-[INFO] after pruning params:  ' + params_new)
    model = model.to(device)
    return model


if __name__ == "__main__":
    from basetrainer.utils import torch_tools
    from torchvision.models.resnet import resnet50, resnet18

    model = resnet18(pretrained=True)
    torch.save(model.state_dict(), "origin.pth")
    model = model_pruning(model, sparsity=0.2)
    torch.save(model.state_dict(), "pruning.pth")



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-07-19 14:19:35
"""
import torch
import torch.nn as nn
import torch.optim as optim


def get_optimizer(model: nn.Module, parameters=[], optim_type="SGD", lr=0.01, split_params=False, **kwargs):
    """
    最终的训练参数是： model.parameters() + parameters
    :param model: nn.Module 模型
    :param parameters: 模型训练参数
    :param optim_type: 优化器, SGD,Adam和AdamW
    :param lr: 学习率
    :param split_params: True: 分离可训练参数，BN层和卷积的bias不加入正则化
                         False: 所以参数，包括BN层和卷积的bias都加入正则化
    :return:
    """
    if split_params:
        params = regularize_parameters(model, weight_decay=kwargs["weight_decay"])
    else:
        params = [{
            'params': list(model.parameters()) + parameters if parameters else list(model.parameters()),
            'weight_decay': kwargs["weight_decay"]
        }]

    if optim_type.lower() == "SGD".lower():
        optimizer = optim.SGD(params, lr=lr, momentum=kwargs["momentum"])
    elif optim_type.lower() == "Adam".lower():
        # β1和β2是加权平均数,用于控制一阶动量和二阶动量
        optimizer = optim.Adam(params, lr=lr)
    elif optim_type.lower() == "AdamW".lower():
        optimizer = optim.AdamW(params, lr=lr)
    else:
        raise Exception("Error:{}".format(optim_type))
    return optimizer


def regularize_parameters(model: nn.Module, weight_decay):
    """
    BatchNorm 层的γ和β不应该添加正则化项,
    卷积层和全连接层的bias也往往不用加正则化项.
    :param model:
    :param optimizer:
    :param weight_decay:
    """
    params_decay = []
    params_no_decay = []
    for m in model.modules():
        if isinstance(m, torch.nn.Linear):
            params_decay.append(m.weight)
            if m.bias is not None:
                params_no_decay.append(m.bias)
        elif isinstance(m, torch.nn.modules.conv._ConvNd):
            params_decay.append(m.weight)
            if m.bias is not None:
                params_no_decay.append(m.bias)
        elif isinstance(m, torch.nn.modules.batchnorm._BatchNorm):
            params_no_decay.extend([*m.parameters()])
        elif len(list(m.children())) == 0:
            params_decay.extend([*m.parameters()])
    assert len(list(model.parameters())) == len(params_decay) + len(params_no_decay)
    return [{'params': params_no_decay}, {'params': params_decay, 'weight_decay': weight_decay}]



# -*-coding: utf-8 -*-
"""
    @Author : PKing
    @E-mail : 390737991@qq.com
    @Date   : 2022-12-26 15:37:56
    @Brief  : https://blog.csdn.net/JianguoChow/article/details/124637324
"""

import os
import torch.onnx
import onnx


def convert2onnx(model, input_shape, input_names=['input'], output_names=['output'],
                 onnx_file="", dynamic=False, simplify=True, opset_version=9,
                 use_prune=False, sparsity=0.2, device="cuda:0"):
    """
    :param model: Pytorch 模型
    :param input_shape: 输入维度(B, C, H, W)
    :param input_names: 输入节点名称
    :param output_names: 输出节点名称
    :param onnx_file: 输出ONNX模型文件
    :param dynamic:
    :param simplify: 是否对ONNX进行simplify
    :param opset_version: ONNX版本
    :param use_prune: 是否对模型进行剪枝
    :param sparsity: 对模型进行剪枝的稀疏度
    :param device: 运行设备
    :return:
    """
    if onnx_file:
        output = os.path.dirname(onnx_file)
    else:
        output = "./output"
        onnx_file = os.path.join(output, "model.onnx")
    if not os.path.exists(output): os.makedirs(output)
    B, C, H, W = input_shape
    if use_prune:
        from basetrainer.pruning import nni_pruning
        model = nni_pruning.model_pruning(model,
                                          input_size=[8, 3, H, W],
                                          sparsity=sparsity,
                                          reuse=True,
                                          output_prune=os.path.join(output, "prune"),
                                          device="cuda")
    model = model.to(device)
    model.eval()
    if dynamic:
        inputs = torch.randn(B, C, H, W).to(device)
        # 声明动态维度，这里我们把input的第0维度赋名为batch_size
        dynamic_axes = {input_names[0]: {0: 'batch_size'}, output_names[0]: {0: 'batch_size'}}
    else:
        inputs = torch.randn(1, C, H, W).to(device)
        dynamic_axes = None
    do_constant_folding = True
    torch.onnx.export(model, inputs,
                      onnx_file,
                      verbose=False,
                      export_params=True,
                      do_constant_folding=do_constant_folding,
                      input_names=input_names,
                      output_names=output_names,
                      dynamic_axes=dynamic_axes,
                      opset_version=opset_version)

    onnx_model = onnx.load(onnx_file)
    onnx.checker.check_model(onnx_model)
    print("save onnx modle file:{}".format(onnx_file))
    # simplify model
    if simplify:
        try:
            import onnxsim
            print(f'simplifying with onnx-simplifier {onnxsim.__version__}')
            onnx_model, check = onnxsim.simplify(onnx_model, dynamic_input_shape=dynamic, input_shapes=None)
            onnx.save(onnx_model, onnx_file)
            print("simplifier onnx model:{}".format(onnx_file))
        except Exception as e:
            print(f'simplifier failure: {e}')
    return onnx_file


if __name__ == "__main__":
    pass



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-07-28 11:32:42
"""
import numpy as np
import torch
from ..metric.eval_tools.metrics import AverageMeter
from ..callbacks import callbacks


class LossesRecorder(callbacks.Callback):
    def __init__(self, indicator="loss"):
        """
        record loss
        :param indicator: 指标名称
        """
        super(LossesRecorder, self).__init__()
        self.indicator = indicator
        self.train_losses = AverageMeter()
        self.test_losses = AverageMeter()

    def on_test_begin(self, logs: dict = {}):
        self.train_losses.reset()
        self.test_losses.reset()

    @staticmethod
    def summary(phase, average_meter: AverageMeter, indicator, losses, logs: dict = {}):
        average_meter.update(losses.data.item())
        logs[phase] = logs[phase] if phase in logs else {}
        logs[phase][indicator] = average_meter.avg

    def on_train_summary(self, inputs, outputs, losses, epoch, step, logs: dict = {}):
        self.summary(phase="train",
                     average_meter=self.train_losses,
                     indicator=self.indicator,
                     losses=losses, logs=logs)

    def on_test_summary(self, inputs, outputs, losses, epoch, batch, logs: dict = {}):
        self.summary(phase="test",
                     average_meter=self.test_losses,
                     indicator=self.indicator,
                     losses=losses, logs=logs)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-12-09 19:16:19
"""
import os
import torch
import torch.nn as nn
import copy
import torch.onnx
from nni.compression.pytorch.utils.counter import count_flops_params
import torch.nn.utils.prune as prune
import torch.nn.functional as F


def model_pruning(model: nn.Module,
                  input_size=[1, 3, 128, 128],
                  sparsity=0.2,
                  prune_mod="L1",
                  reuse=False,
                  speedup=True,
                  output_prune="pruning_output",
                  mask_file="",
                  dependency_aware=True,
                  device="cpu",
                  verbose=False,
                  **kwargs):
    """
    使用torch.nn.utils.prune进行模型剪枝和压缩
    https://pytorch.org/tutorials/intermediate/pruning_tutorial.html?highlight=prune
    :param model:  Pytorch模型
    :param input_size: 模型输入的维度[batch-size,channel,H,W]
    :param sparsity: 模型剪枝目标稀疏度,值越大,模型越稀疏，计算量越小；0.5表示剪除50%
    :param reuse: 是否复现模型剪枝的结果
                False: 进行模型剪枝(pruning+ModelSpeedup),会生成mask文件(mask_naive_l1filter.pth)
                True : 复现模型剪枝的结果,需要提供mask_file文件
    :param speedup: 是否加速模型
    :param output_prune: 模型剪枝输出文件
    :param mask_file: reuse=True需要提供模型剪枝的mask文件,默认保存在output_prune目录下(mask_naive_l1filter.pth)
    :param dependency_aware 依赖感知模式 https://nni.readthedocs.io/zh/stable/Compression/DependencyAware.html
    :param device:
    :return:
    """
    info = ""
    model = model.to(device)
    if not os.path.exists(output_prune): os.makedirs(output_prune)
    prune_file = os.path.join(output_prune, 'pruned_naive_{}filter.pth'.format(prune_mod))
    onnx_file = os.path.join(output_prune, 'pruned_naive_{}filter.onnx'.format(prune_mod))
    mask_file = os.path.join(output_prune, 'mask_naive_{}filter.pth'.format(prune_mod)) if not mask_file else mask_file
    dummy_input = torch.randn(input_size).to(device)
    # 原始模型的计算量和参数量
    flops, params, _ = count_flops_params(model, dummy_input, verbose=verbose)
    info += f"origin-Model FLOPs {flops / 1e6:.2f}M, Params {params / 1e6:.2f}M\n"
    # 定义需要剪枝的参数
    parameters_to_prune = []
    for name, module in model.named_modules():
        if isinstance(module, torch.nn.Conv2d):
            # parameters += [(module, 'weight'), (module, 'bias')]
            parameters_to_prune += [(module, 'weight')]
            if hasattr(module, 'bias') and module.bias is not None:
                parameters_to_prune += [(module, 'bias')]
        if isinstance(module, torch.nn.Linear):
            # parameters += [(module, 'weight'), (module, 'bias')]
            parameters_to_prune += [(module, 'weight')]
            if hasattr(module, 'bias') and module.bias is not None:
                parameters_to_prune += [(module, 'bias')]
    # 全局剪枝
    prune.global_unstructured(parameters_to_prune,
                              pruning_method=prune.L1Unstructured,
                              amount=sparsity, )
    # to verify that all masks exist
    # print(dict(model.named_buffers()).keys())
    # apply_compression_results_v1(prune, model)
    # 剪枝模型的计算量和参数量
    flops, params, _ = count_flops_params(model, dummy_input, verbose=verbose)
    info += f"prune-Model FLOPs {flops / 1e6:.2f}M, Params {params / 1e6:.2f}M\n"
    print(info)
    return model


def apply_compression_results_v1(prune: torch.nn.utils.prune, model: nn.Module):
    # Apply mask to model tensor and remove mask from state_dict
    for name, module in model.named_modules():
        if isinstance(module, torch.nn.Conv2d):
            prune.remove(module, 'weight')
            if hasattr(module, 'bias') and module.bias is not None:
                prune.remove(module, 'bias')
        if isinstance(module, torch.nn.Linear):
            prune.remove(module, 'weight')
            if hasattr(module, 'bias') and module.bias is not None:
                prune.remove(module, 'bias')


def apply_compression_results_v2(model):
    """
    Apply the masks from ```masks_file``` to the model
    Note: this API is for inference, because it simply multiplies weights with
    corresponding masks when this API is called.
    model : torch.nn.ModuleThe model to be compressed
    """
    for name, module in model.named_modules():
        if isinstance(module, torch.nn.Conv2d):
            module.weight.data = module.weight.data.mul_(module.weight_mask)
            if hasattr(module, 'bias') and module.bias is not None:
                module.bias.data = module.bias.data.mul_(module.bias_mask)
        if isinstance(module, torch.nn.Linear):
            module.weight.data = module.weight.data.mul_(module.weight_mask)
            if hasattr(module, 'bias') and module.bias is not None:
                module.bias.data = module.bias.data.mul_(module.bias_mask)


class SimpleModel(nn.Module):
    def __init__(self):
        super(SimpleModel, self).__init__()
        self.conv1 = nn.Conv2d(3, 32, 3)
        self.conv2 = nn.Conv2d(32, 64, 3)
        self.conv3 = nn.Conv2d(64, 128, 3)
        self.fc = nn.Linear(128, 256)
        self.classifier = nn.Linear(256, 10)

    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        x = F.relu(self.conv3(x))
        x = nn.functional.adaptive_avg_pool2d(x, 1).reshape(x.shape[0], -1)
        x = self.fc(x)
        x = self.classifier(x)
        return x


if __name__ == "__main__":
    from torchvision.models.resnet import resnet50, resnet18
    from torchvision.models.squeezenet import SqueezeNet
    from torchvision.models.mobilenet import MobileNetV2
    from segment.models import build_model
    from libs.performance import performance

    device = "cuda:0"
    batch_size = 2
    width_mult = 1.0
    num_classes = 20
    input_size = [1, 3, 224, 224]
    net_type = 'modnet_v2'
    model = resnet18(pretrained=True)
    # model = SimpleModel()
    # model = MobileNetV2()
    # model = build_model.get_models(net_type, input_size, num_classes, width_mult=width_mult, pretrained=False)
    model.eval()
    inputs = torch.randn(input_size)
    model = model.to((device))
    inputs = inputs.to((device))
    out = model(inputs)
    performance.model_performance(model, inputs)
    prune_model = copy.deepcopy(model)
    prune_model = model_pruning(prune_model, input_size=input_size, sparsity=0.9, dependency_aware=True, device=device)
    performance.model_performance(model, inputs)
    performance.model_performance(prune_model, inputs)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-01-25 18:39:09
"""
import numpy as np


class MultiStepValue(object):
    def __init__(self,
                 epochs,
                 steps_per_epoch,
                 milestones,
                 val_list=[],
                 val_init=0.01,
                 val_end=0,
                 num_warn_up=0,
                 decay_rates=None):
        """
        :param epochs:
        :param steps_per_epoch:
        :param lr_min : lr_min
        :param val_init: lr_max is init lr.
        :param num_warn_up:
        """
        super(MultiStepValue, self).__init__()
        self.epochs = epochs
        self.steps_per_epoch = steps_per_epoch
        self.max_step = epochs * self.steps_per_epoch
        self.val_init = val_init
        self.val_end = val_end
        self.milestones = milestones
        self.warmup = num_warn_up * self.steps_per_epoch
        if not val_list:
            if decay_rates:
                val_list = [val_init * decay for decay in decay_rates]
            else:
                val_list = [val_init * 0.1 ** decay for decay in range(0, len(self.milestones) + 1)]
            if val_end is not None:
                val_list[-1] = val_end
        self.val_list = val_list
        self.val = self.val_init
        print("val_list:{}".format(self.val_list))
        print("milestones:{}".format(self.milestones))
        print("val_init:{}".format(self.val_init))
        assert len(self.milestones) == len(self.val_list) - 1

    def get_val(self):
        return self.val

    def __get_val(self, epoch, lr_stages, lr_list):
        """
        :param epoch:
        :param lr_stages:
        :param lr_list:
        :return:
        """
        lr = None
        max_stages = 0
        if not lr_stages:
            lr = lr_list[0]
        else:
            max_stages = max(lr_stages)
        for index in range(len(lr_stages)):
            if epoch < lr_stages[index]:
                lr = lr_list[index]
                break
            if epoch >= max_stages:
                lr = lr_list[index + 1]
        return lr

    def __set_stages_val(self, epoch, lr_stages, lr_list):
        '''
        :param epoch:
        :param lr_stages: [    35, 65, 95, 150]
        :param lr_list:   [0.1, 0.01, 0.001, 0.0001, 0.00001]
        :return:
        '''
        lr = self.__get_val(epoch, lr_stages, lr_list)
        if lr is not None:
            self.val = lr

    def set_val(self, epoch, total_step):
        """
        Usage:
        for epoch in range(epochs):
            for i in range(steps_per_epoch):
                scheduler.on_step(steps_per_epoch * epoch + i)
                ...
        :param total_step: total step: steps_per_epoch * epoch + step
        :return:
        """
        if self.warmup and total_step <= self.warmup:
            self.val = self.val_init / self.warmup * total_step
        else:
            self.__set_stages_val(epoch, self.milestones, self.val_list)

    def step(self, epoch, step=0):
        total_step = self.steps_per_epoch * epoch + step
        self.set_val(epoch, total_step)


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    # Plot lr schedule
    epochs = 200
    steps_per_epoch = 100
    num_warn_up = 0
    milestones = list(range(15, epochs - 15, 15))
    val_list = np.arange(1, 0, -1 / len(milestones)).tolist() + [0]
    val_init = 1.0

    kd_decay_scheduler = MultiStepValue(epochs, steps_per_epoch,
                                        val_list=val_list,
                                        val_init=val_init,
                                        milestones=milestones,
                                        num_warn_up=num_warn_up)
    lr_steps = []
    lr_epochs = []
    for epoch in range(epochs):
        # scheduler.epoch(epoch)
        # for i in range(steps_per_epoch):
        kd_decay_scheduler.step(epoch)
        lr = kd_decay_scheduler.get_val()
        lr_steps.append(lr)
        lr_epochs.append(lr)
        print(epoch, lr)
    plt.figure()
    plt.plot(lr_steps, label="lr_steps")
    plt.grid(True)  # 显示网格;
    plt.xlabel("steps")
    plt.ylabel("LR")

    plt.figure()
    plt.plot(lr_epochs, label="lr_epochs")
    plt.grid(True)  # 显示网格;
    plt.xlabel("epochd")
    plt.ylabel("LR")

    plt.show()



# -*- coding: utf-8 -*-
"""
# --------------------------------------------------------
# @Project: torch-Face-Recognize-Pipeline
# @Author : panjq
# @E-mail : pan_jinquan@163.com
# @Date   : 2019-12-31 09:11:25
# --------------------------------------------------------
"""
import numpy as np
import torch
import torch.nn as nn
from typing import Dict, List, Callable


def get_criterion(loss_type: str, num_classes=None, class_weight=None, device="cuda:0", **kwargs):
    """
    :param loss_type: loss_type={loss_name: loss_weigth}
                      FocalLoss,CrossEntropyLoss,LabelSmooth
    :param num_classes:
    :param class_weight: 类别loss权重， a manual rescaling weight given to each class.
                         If given, has to be a Tensor of size `Class`
    :return:
    """
    loss_type = loss_type.lower()
    if loss_type == "CrossEntropyLoss".lower() or loss_type == "CELoss".lower():
        criterion = nn.CrossEntropyLoss(weight=class_weight)
    elif loss_type == "L1Loss".lower():
        criterion = nn.L1Loss(reduction='mean')
    elif loss_type == "mse".lower():
        # loss = nn.MSELoss(reduce=True, size_average=True)
        criterion = nn.MSELoss(reduction='mean')
    else:
        raise Exception("Error:{}".format(loss_type))
    print("loss_type:{}".format(loss_type))
    return criterion


def build_criterion(loss_type: str or List[str] or Dict[str, float],
                    num_classes=None,
                    class_weight=None,
                    device="cuda:0",
                    **kwargs):
    """
    使用nn.BCELoss需要在该层前面加上Sigmoid函数
    使用nn.CrossEntropyLoss会自动加上Softmax层,所以最后一层不需要加上Softmax()
    :param loss_type: loss_type={loss_name: loss_weigth}
                      FocalLoss,CrossEntropyLoss,LabelSmooth
    :param num_classes:
    :param class_weight: 类别loss权重， a manual rescaling weight given to each class.
                         If given, has to be a Tensor of size `Class`
    :return:
    """
    # logger = log.get_logger()
    if isinstance(class_weight, np.ndarray):
        class_weight = torch.from_numpy(class_weight.astype(np.float32)).to(device)
    if isinstance(loss_type, str):
        loss_type = [loss_type]
    if isinstance(loss_type, list):
        loss_type = {loss: 1.0 for loss in loss_type}
    assert isinstance(loss_type, dict)
    criterions = {}
    weights = {}
    for loss, loss_weight in loss_type.items():
        criterion = get_criterion(loss, num_classes, class_weight, device=device, **kwargs)
        criterions[loss] = criterion
        weights[loss] = loss_weight
    criterions = ComposeLoss(criterions=criterions, weights=weights)
    # logger.info("use criterions:{}".format(weights))
    return criterions


class ComposeLoss(nn.Module):
    def __init__(self, criterions: Dict[str, Callable], weights: Dict[str, float] = None):
        """
        联合LOSS函数
        :param criterions: Dict[str, Callable], Loss函数,
                         ==> {"Loss1": Loss1Function,"Loss2":Loss2Function}
        :param  weights: Dict[str, float] Loss的权重
                         ==> {"Loss1": 1.0,"Loss2":1.0}
        """
        super(ComposeLoss, self).__init__()
        if isinstance(weights, dict):
            assert criterions.keys() == weights.keys(), \
                Exception("Key Error:criterions:{},weights:{}".format(criterions.keys(), weights.keys()))
        self.weights = weights
        self.criterions = criterions

    def parameters(self, recurse: bool = True):
        param = []
        for loss in self.criterions.values():
            param += loss.parameters()
        return param

    def __call__(self, logits, labels):
        losses = {}
        for name, criterion in self.criterions.items():
            loss = criterion(logits, labels)
            if isinstance(loss, dict):
                losses = {k: v * self.weights[name] for k, v in loss.items()} if self.weights else loss
            else:
                losses[name] = loss * self.weights[name] if self.weights else loss
        return losses


if __name__ == "__main__":
    from basetrainer.utils import torch_tools

    torch_tools.set_env_random_seed()
    batch_size = 16
    num_classes = 10
    input = torch.ones(batch_size, num_classes, 10, 10).cuda()
    target = torch.ones(batch_size, 10, 10).long().cuda()
    loss_type = {"CELoss": 0.5, "CELoss": 0.5}
    losses = build_criterion(loss_type, num_classes=num_classes, device="cuda:0")
    loss = losses(input, target)
    print(loss)



# -*-coding: utf-8 -*-
"""
    @Project: torch-Face-Recognize-Pipeline
    @File   : evaluate.py
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2019-12-26 16:32:47
"""

import torch

from evaluation.eval_tools.metrics import AverageMeter, accuracy


class EvaluationMetrics(object):
    def __init__(self, model=None, device="cuda:0"):
        self.model = model
        self.device = device

    def update_model(self, model, gpu_id):
        """
        :param model:
        :param gpu_id:
        :return:
        """
        if len(gpu_id) > 1:
            model = model.module
            self.model = model.to(self.device)
        else:
            self.model = model.to(self.device)
        # switch to evaluation mode
        self.model.eval()

    def forward(self, input_tensor):
        """
        :param input_tensor: input tensor
        :return:
        """
        with torch.no_grad():
            out_tensor = self.model(input_tensor.to(self.device))
        return out_tensor


    def evaluation(self,inputs,labels, epoch):
        """
        val data metrics
        :param epoch:
        :return:
        """
        self.model.eval()  # set to training mode
        losses = AverageMeter()
        top1 = AverageMeter()
        inputs = inputs.to(self.device)
        labels = labels.to(self.device)
        outputs = self.model(inputs)
        loss = self.loss(outputs, labels)
        # measure accuracy and record loss
        acc, = accuracy(outputs.data, labels, topk=(1,))
        losses.update(loss.data.item(), inputs.size(0))
        top1.update(acc.data.item(), inputs.size(0))



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-08-13 09:13:03
"""
import torch
import torch.optim as optim
import torch.nn as nn
import numpy as np
from ..callbacks.callbacks import Callback
from torch.optim import lr_scheduler


class WarmUpLR(Callback):
    """warmup_training learning rate scheduler
    Args:
        optimizer: optimzier(e.g. SGD)
        total_iters: totoal_iters of warmup phase
    """

    def __init__(self,
                 optimizer,
                 num_steps,
                 lr_init=0.01,
                 num_warn_up=0):
        super(WarmUpLR, self).__init__()
        self.optimizer = optimizer
        self.num_steps = num_steps
        self.lr_init = lr_init
        self.warmup = num_warn_up * self.num_steps
        self.epoch = 0
        super(WarmUpLR, self).__init__()

    def set_lr(self, lr):
        for param_group in self.optimizer.param_groups:
            param_group["lr"] = lr

    def get_lr(self, total_step):
        """we will use the first m batches, and set the learning
        rate to base_lr * m / total_iters
        """
        lr = self.lr_init / self.warmup * total_step
        # lr= self.optimizer.param_groups[0]["lr"]
        return lr

    def on_epoch_begin(self, epoch, logs: dict = {}):
        self.epoch = epoch

    def on_batch_end(self, batch, logs: dict = {}):
        self.step(epoch=self.epoch, step=batch)

    def step(self, epoch=0, step=0):
        total_step = self.num_steps * epoch + step
        if self.warmup > 0 and total_step <= self.warmup:
            lr = self.get_lr(total_step)
            self.set_lr(lr)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-07-02 14:59:42
"""
import os
import sys
import tensorboardX as tensorboard
# from torch.utils import tensorboard


class SummaryWriter():
    def __init__(self, log_dir, *args, **kwargs):
        self.tensorboard = None
        if log_dir:
            # 修复tensorboard版本BUG
            self.tensorboard = tensorboard.SummaryWriter(log_dir, *args, **kwargs)

    def add_scalar(self, *args, **kwargs):
        if self.tensorboard:
            self.tensorboard.add_scalar(*args, **kwargs)

    def add_scalars(self, *args, **kwargs):
        if self.tensorboard:
            self.tensorboard.add_scalars(*args, **kwargs)

    def add_image(self, *args, **kwargs):
        if self.tensorboard:
            self.tensorboard.add_image(*args, **kwargs)


if __name__ == '__main__':
    main_process = True
    log_root = "./"
    writer1 = SummaryWriter(log_root if main_process else None)
    # main_process=False
    writer2 = SummaryWriter(log_root if main_process else None)
    writer3 = SummaryWriter(log_root if main_process else None)
    epochs = 200
    for epoch in range(epochs):
        if writer3:
            print(writer3)
        writer1.add_scalar("lr_epoch", epoch, epoch)
        writer2.add_scalar("lr_epoch", epoch, epoch)
        writer3.add_scalar("lr_epoch", epoch, epoch)



# -*-coding: utf-8 -*-

import os, sys

sys.path.append(os.getcwd())
import onnxruntime
import onnx


class ONNXEngine():
    def __init__(self, onnx_path):
        """
        :param onnx_path:
        """
        self.onnx_session = onnxruntime.InferenceSession(onnx_path)
        self.input_name = self.get_input_name(self.onnx_session)
        self.output_name = self.get_output_name(self.onnx_session)
        print("input_name:{}".format(self.input_name))
        print("output_name:{}".format(self.output_name))

    def get_output_name(self, onnx_session):
        """
        output_name = onnx_session.get_outputs()[0].name
        :param onnx_session:
        :return:
        """
        output_name = []
        for node in onnx_session.get_outputs():
            output_name.append(node.name)
        return output_name

    def get_input_name(self, onnx_session):
        """
        input_name = onnx_session.get_inputs()[0].name
        :param onnx_session:
        :return:
        """
        input_name = []
        for node in onnx_session.get_inputs():
            input_name.append(node.name)
        return input_name

    def get_input_feed(self, input_name, image_tensor):
        """
        input_feed={self.input_name: image_tensor}
        :param input_name:
        :param image_tensor:
        :return:
        """
        input_feed = {}
        for name in input_name:
            input_feed[name] = image_tensor
        return input_feed

    def __call__(self, image_tensor):
        """
        num_bboxes=1500
        num_class=1
        :param image_tensor: shape
        :return: outputs[0]
        """
        outputs = self.forward(image_tensor)
        return outputs

    def forward(self, image_tensor):
        """
        image_tensor = image.transpose(2, 0, 1)
        image_tensor = image_tensor[np.newaxis, :]
        onnx_session.run([output_name], {input_name: x})
        :param image_tensor:
        :return:
        """
        # 输入数据的类型必须与模型一致,以下三种写法都是可以的
        # scores, boxes = self.onnx_session.run(None, {self.input_name: image_tensor})
        # scores, boxes = self.onnx_session.run(self.output_name, input_feed={self.input_name: image_tensor})
        input_feed = self.get_input_feed(self.input_name, image_tensor)
        outputs = self.onnx_session.run(self.output_name, input_feed=input_feed)
        return outputs


if __name__ == "__main__":
    import numpy as np

    model_file = "/home/dm/data3/release/infrastructure/DMClassification/work_space/test/mobilenet_v2_1.0_128_128_CrossEntropyLoss_SGD_20211210123127/model/best_model_052_99.4845.onnx"
    # pmodel_file = "/home/dm/data3/release/infrastructure/DMClassification/classifier/models/models/slim_pruned.onnx"
    prune_file = "/home/dm/data3/release/infrastructure/DMClassification/work_space/test/mobilenet_v2_1.0_128_128_CrossEntropyLoss_SGD_20211210141358/model/best_model_008_97.9381.onnx"
    batch_size = 1
    num_classes = 4
    input_size = [128, 128]
    inputs = np.random.random(size=(batch_size, 3, input_size[1], input_size[0]))
    inputs = np.asarray(inputs, dtype=np.float32)
    model = ONNXEngine(model_file)
    print("----")



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-07-30 17:04:51
"""

import torch.nn as nn
import torch.utils.data as torch_utils
from torch.utils.data.dataset import Dataset
from torch.utils.data.dataloader import DataLoader
from ..engine import comm
from .torch_tools import get_torch_version, torch_version_id


def build_dataloader(dataset: Dataset,
                     batch_size: int,
                     num_workers: int,
                     shuffle: bool = True,
                     persistent_workers: bool = False,
                     phase: str = "train",
                     distributed=True,
                     **kwargs) -> DataLoader:
    """
    :param dataset: Dataset
    :param batch_size: DP模式中，每张卡将分配(batch_size/GPU)数进行训练；而DDP模式中，每张卡都有batch_size的数据
    :param num_workers:
    :param shuffle: 是否打乱顺序
    :param persistent_workers: 该参数仅支持torch>=1.6
           False: 数据加载器运行完一个Epoch后会关闭worker进程,在分布式训练，会出现每个epoch初始化多进程的问题
           True: 会保持worker进程实例激活状态,容易出现“DataLoader worker (pid(s) 953) exited unexpectedly”的错误
    :param phase: "train", "test", "val"
    :param distributed: True: use DDP; False: use DP (是否使用分布式训练)
    :param kwargs:
    :return:
    """
    assert phase in ["train", "test", "val"]
    sampler = None
    if comm.get_world_size() > 1 and phase == "train" and distributed:
        # DistributedSampler为每个子进程分发数据，避免数据重复
        sampler = torch_utils.distributed.DistributedSampler(dataset,
                                                             num_replicas=comm.get_world_size(),
                                                             rank=comm.get_local_rank(),
                                                             shuffle=shuffle)
        shuffle = False  # sampler option is mutually exclusive with shuffle
    else:
        # Fix a Bug: RuntimeError: can't start new thread
        # Fix a Bug: DataLoader worker (pid(s) 953) exited unexpectedly
        persistent_workers = False
    try:
        # Fix a Bug: torch<=1.6 have no argument 'persistent_workers'
        if get_torch_version() >= torch_version_id("1.7"):
            kwargs["persistent_workers"] = persistent_workers
            # fix a bug: persistent_workers option needs num_workers > 0
            if persistent_workers and num_workers == 0:
                kwargs["persistent_workers"] = False
    except:
        print("torch<=1.6 have no argument persistent_workers")
    dataloader = torch_utils.DataLoader(dataset,
                                        batch_size=batch_size,
                                        num_workers=num_workers,
                                        sampler=sampler,
                                        shuffle=shuffle,
                                        **kwargs)
    return dataloader


def build_model_parallel(model: nn.Module,
                         device_ids=None,
                         distributed=True,
                         **kwargs) -> nn.Module:
    """
    :param model:
    :param device_ids:
    :param distributed: True: use DDP; False: use DP (是否使用分布式训练)
    :param kwargs:
    :return:
    """
    print("device_ids:{},device:{}".format(device_ids, comm.get_device(device_ids)))
    model.to(comm.get_device(device_ids))
    # use DistributedDataParallel
    if comm.get_world_size() > 1 and distributed:
        model = nn.parallel.DistributedDataParallel(model,
                                                    device_ids=[comm.get_device(device_ids)],
                                                    output_device=comm.get_device(device_ids),
                                                    **kwargs
                                                    )
    else:
        # use DataParallel
        model = nn.DataParallel(model, device_ids=device_ids, output_device=comm.get_device(device_ids), **kwargs)
    return model
